from dataclasses import dataclass

import numpy as np
import pytest
from typing import Any, Dict

from .. import log
from ..consensus import DiscreteConsensusProblem
from ..problems import JSONDataClass


@pytest.fixture(scope="module")
def dcp_kwargs() -> Dict[str, Any]:
    return {
        "n_tasks": 2,
        "n_workers": 2,
        "t_A": [0, 1, 1],
        "w_A": [0, 0, 1],
        "n_labels": 2,
        "f_A": [0, 0, 1]}


@pytest.fixture(scope="module")
def manual_n() -> np.ndarray:
    """Hand-computed `n` array from `dcp_kwargs`"""
    return np.array([[[1, 0],
                      [1, 0]],
                     [[0, 0],
                      [0, 1]]])


@pytest.fixture(scope="module")
def dcp_kwargs_2() -> Dict[str, Any]:
    return {
        "n_tasks": 2,
        "n_workers": 2,
        "t_A": [1, 1, 1],  # task 0 has no annotations
        "w_A": [0, 0, 1],
        "n_labels": 2,
        "f_A": [0, 0, 1]}


@pytest.fixture(scope="module")
def manual_n2_filtered() -> np.ndarray:
    """Hand-computed `n` array from `dcp_kwargs_2` w/ `ignore_zero_annots=True`"""
    return np.array([[[2, 0]],  # task 0 is filtered
                     [[0, 1]]])


@pytest.fixture(scope="module")
def manual_n2_unfiltered() -> np.ndarray:
    """Hand-computed `n` array from `dcp_kwargs_2` w/ `ignore_zero_annots=False`"""
    return np.array([[[0, 0],
                      [2, 0]],
                     [[0, 0],
                      [0, 1]]])


@pytest.fixture(scope="module")
def dcp(dcp_kwargs) -> DiscreteConsensusProblem:
    dcp_ = DiscreteConsensusProblem(**dcp_kwargs)
    log.info("dcp from kwargs: {}".format(dcp_))
    log.info("dcp.to_json(): {}".format(dcp_.to_json()))
    return dcp_


def test_constructors(dcp_kwargs, dcp):
    assert dcp.n_tasks == dcp_kwargs["n_tasks"]
    assert np.array_equal(dcp.t_A, dcp_kwargs["t_A"])
    assert dcp.f_T is None
    kw = dcp_kwargs.copy()
    kw["n_tasks"] = dcp_kwargs["n_tasks"] * 2
    assert not dcp == DiscreteConsensusProblem(**kw)


def test_json_methods(dcp):
    assert dcp == DiscreteConsensusProblem.from_json(dcp.to_json())


def test_compute_n(dcp, manual_n, dcp_kwargs_2, manual_n2_filtered, manual_n2_unfiltered):
    n, _ = dcp.compute_n()
    assert (dcp.n_workers, dcp.n_tasks, dcp.n_labels) == n.shape
    assert np.array_equal(n, manual_n)
    assert np.array_equal(dcp.compute_n(ignore_zero_annots=True)[0], dcp.compute_n(ignore_zero_annots=False)[0])

    dcp2 = DiscreteConsensusProblem(**dcp_kwargs_2)
    n2_unfiltered, _ = dcp2.compute_n(ignore_zero_annots=False)
    assert np.array_equal(n2_unfiltered, manual_n2_unfiltered)
    n2_filtered, filtered_tasks = dcp2.compute_n(ignore_zero_annots=True)
    assert np.array_equal(n2_filtered, manual_n2_filtered)
    assert np.array_equal(filtered_tasks, np.array([0]))  # task 0 was filtered


def test_jsondataclass_product_from_dict():
    @dataclass
    class C(JSONDataClass):
        a: int
        b: int
    options = {"a": [10, 100], "b": [3, 9]}
    assert C.product_from_dict(options) == [C(10, 3), C(10, 9), C(100, 3), C(100, 9)]





import ast
import dataclasses
from .. import log
from importlib.resources import path as irpath
from typing import Dict, Tuple

import numpy as np
from cmdstanpy import CmdStanModel, CmdStanMLE

from ..consensus import GenerativeAbstractConsensus, DiscreteConsensusProblem


def _parse_var_dims_and_index_ranges(names: Tuple[str, ...]) -> Dict:
    """
    Use Stan CSV file column names to get variable names, dimensions.
    Assumes that CSV file has been validated and column names are correct.
    """
    if names is None:
        raise ValueError('missing argument "names"')
    vars_dict = {}
    idx = 0
    while idx < len(names):
        if names[idx].endswith('__'):
            pass
        elif '[' not in names[idx]:
            vars_dict[names[idx]] = (1, slice(idx))
        else:
            vs = names[idx].split('[')
            idx_start = idx
            while idx < len(names) - 1 and names[idx + 1].split('[')[0] == vs[0]:
                idx += 1
            vs = names[idx].split('[')
            dims = [int(x) for x in vs[1][:-1].split(',')]
            vars_dict[vs[0]] = (tuple(dims), slice(idx_start, idx + 1))
        idx += 1
    return vars_dict


def _get_var_dict(results: CmdStanMLE):
    var_info = _parse_var_dims_and_index_ranges(results.column_names)
    values_array = results.optimized_params_np
    var_dict = {var_name: values_array[var_info[var_name][1]].reshape(var_info[var_name][0], order='F')
                for var_name in var_info
                }
    return var_dict


with irpath(__package__, "__init__.py") as path_:
    resource_dir = __name__.split('.')[-1]  # directory name for the *.stan files
    resources_path = path_.parents[0] / resource_dir


def resource_filename(filename):
    return resources_path / filename


class AbstractStanOptimizeConsensus(GenerativeAbstractConsensus):
    name = "AbstractStanOptimizeConsensus"

    def __init__(self, model_name):
        self.model_name = model_name
        # self.hidden_variables = []

    def map_data_to_model(self, dcp: DiscreteConsensusProblem, **kwargs):
        d = self.map_data_to_data(dcp, **kwargs)
        init = self.map_data_to_inits(dcp, **d, **kwargs)
        args = self.map_data_to_args(dcp, **d, **kwargs)
        return d, init, args

    def map_data_to_data(self, dcp: DiscreteConsensusProblem, **kwargs):
        stan_data = {'w': dcp.n_workers,
                     't': dcp.n_tasks,
                     'a': dcp.n_annotations,
                     'k': len(dcp.classes),
                     'l': dcp.n_labels,
                     'classes': (np.array(dcp.classes) + 1),
                     't_A': (dcp.t_A + 1),
                     'w_A': (dcp.w_A + 1),
                     'ann': (dcp.f_A + 1).flatten()}
        prior = self.map_data_to_prior(dcp, **stan_data, **kwargs)
        log.debug(prior)
        stan_data.update(prior)
        return stan_data

    def map_data_to_prior(self, dcp, **kwargs):
        raise NotImplementedError

    def map_data_to_inits(self, dcp,  **kwargs):
        raise NotImplementedError

    def map_data_to_args(self, dcp, **kwargs):
        raise NotImplementedError

    # def MLE_parameters(self, results: CmdStanMLE):
    #     return {var_name: results.optimized_params_dict[var_name] for var_name in self.hidden_variables}

    def fit_and_compute_consensus_model(self):
        return CmdStanModel(stan_file=resource_filename(self.model_name + ".fit_and_consensus.stan"))

    def fit_and_compute_consensus(self, dcp: DiscreteConsensusProblem, **kwargs):
        """Fits the model parameters and computes the consensus.
        returns consensus, model parameters"""

        stan_data, init_data, kwargs = self.map_data_to_model(dcp)
        model = self.fit_and_compute_consensus_model()
        log.info(stan_data.keys())
        for f in stan_data.keys():
            log.info("Type of %s is %s", f, type(stan_data[f]))
        log.info(init_data)
        # stan_data = {}
        # init_data = {}
        results = model.optimize(data=stan_data, inits=init_data, **kwargs)
        var_dict = _get_var_dict(results)
        keys = [x.name for x in dataclasses.fields(self.Parameters)]
        filtered_dict = {x: var_dict[x] for x in keys}
        return var_dict["t_C"], self.Parameters(**filtered_dict)

    def fit_model(self):
        return CmdStanModel(stan_file=resource_filename(self.model_name + ".fit.stan"))

    def fit(self, dcp: DiscreteConsensusProblem, reference_consensus, **kwargs):
        """ Fits the model parameters provided that the consensus is already known.
        This is useful to determine the errors of a different set of annotators than the
        ones that were used to determine the consensus.
        """

        stan_data, init_data, kwargs = self.map_data_to_model(dcp)
        stan_data['t_C'] = reference_consensus

        model = self.fit_model()
        results = model.optimize(data=stan_data, inits=init_data, **kwargs)
        keys = [x.name for x in dataclasses.fields(self.Parameters)]
        d_results = _get_var_dict(results)
        filtered_dict = {x: d_results[x] for x in keys}
        return self.Parameters(**filtered_dict)

    def compute_consensus_model(self):
        return CmdStanModel(stan_file=resource_filename(self.model_name + ".consensus.stan"))

    def compute_consensus(self, dcp: DiscreteConsensusProblem, parameters: GenerativeAbstractConsensus.Parameters,
                          **kwargs):
        # print(kwargs["data"])
        stan_data, init_data, kwargs_opt = self.map_data_to_model(dcp)
        # print("stan_data:", stan_data)
        # print("kwargs['data']:", ast.literal_eval(kwargs["data"]))
        # stan_data.update(ast.literal_eval(kwargs["data"]))
        stan_data.update(ast.literal_eval(parameters.to_json()))

        model = self.compute_consensus_model()
        results = model.optimize(data=stan_data, inits=init_data, **kwargs_opt)

        return _get_var_dict(results)['t_C'], kwargs_opt

    def sample_tasks_model(self):
        return CmdStanModel(stan_file=resource_filename(self.model_name + ".sample_tasks.stan"))

    def sample_annotations_model(self):
        return CmdStanModel(stan_file=resource_filename(self.model_name + ".sample_annotations.stan"))



import dataclasses
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Tuple, Type

import pytest
import numpy as np

from . import close, distance, TOLERANCE
from .. import log
from ..consensus import AbstractSimpleConsensus, GenerativeAbstractConsensus
from ..measures import Accuracy, AllClose
from ..problems import ConsensusProblem


Kwargs = Dict[str, Any]


@dataclass
class SampleForTest:
    """Class for a sampling function's output"""
    problem: ConsensusProblem
    parameters: GenerativeAbstractConsensus.Parameters
    kwargs: Kwargs = dataclasses.field(default_factory=dict)  # Intended for `fit_and_compute_consensus` only


class BaseTestSimpleConsensusModel:
    """Abstract test class for simple consensus models.

        Conducts following tests:
        1. `model_cls` class attribute is valid;
        2. `sampling_funcs` class attribute is valid;
        3. Asserts each sample provided by sampling functions is valid;
        4. Runs tests for the following methods of the child consensus model:
            * `fit_and_compute_consensus`

        Notes:
            A child test class only has to override `model_cls` and `sampling_funcs` class attributes.
    """
    # Class of the model to be tested; e.g. MajorityVoting
    model_cls: Type[AbstractSimpleConsensus] = None
    # A list of zero-argument sampling functions that return `SampleTest` objects
    sampling_funcs: List[Callable[[], SampleForTest]] = []
    # Absolute tolerance for the `numpy.allclose` function that may also be overridden, if need be
    ABSOLUTE_TOLERANCE = TOLERANCE

    @classmethod
    @pytest.fixture(scope="class")
    def samples(cls) -> List[SampleForTest]:
        # print("In samples for the model {}".format(cls.model_cls.name))
        return [sf() for sf in cls.sampling_funcs]

    @classmethod
    @pytest.fixture(scope="class")
    def ref_consensus_params_problem(cls, samples) -> List[Tuple[np.ndarray,
                                                                 GenerativeAbstractConsensus.Parameters,
                                                                 ConsensusProblem,
                                                                 Kwargs]]:
        # print("In ref_consensus_params_problem for the model {}".format(cls.model_cls.name))
        model = cls.model_cls()
        list_ref = []
        for sample in samples:
            consensus_ref, parameters_ref = model.fit_and_compute_consensus(sample.problem)
            list_ref.append((consensus_ref, parameters_ref, sample.problem, sample.kwargs))
        return list_ref

    @classmethod
    def _test_fit_and_compute_consensus(cls, sample: SampleForTest):
        model = cls.model_cls()
        # print(sample.problem)
        consensus, parameters_learned = model.fit_and_compute_consensus(sample.problem, **sample.kwargs)
        # Assert consensus
        if cls.model_cls.__bases__[0] == AbstractSimpleConsensus:  # If this is only a simple consensus model
            assert Accuracy.evaluate(sample.problem.f_T, consensus) == 1.0
        else:
            assert 1. - Accuracy.evaluate(sample.problem.f_T, consensus) <= cls.ABSOLUTE_TOLERANCE
        # Assert parameters
        if sample.parameters:
            dict_parameters_learned = dataclasses.asdict(parameters_learned)
            dict_sample_parameters = dataclasses.asdict(sample.parameters)
            for p in dict_sample_parameters.keys():
                log.debug("Distance between learned {p} and real {p}: {d:f}".format(
                    p=p, d=distance(dict_parameters_learned[p], dict_sample_parameters[p])))
                assert close(dict_parameters_learned[p], dict_sample_parameters[p])

    # ==============================  Actual tests  ==============================
    def test_model_cls(self):
        # print("In test_model_cls for the model {}".format(self.model_cls.name))
        assert self.model_cls
        assert issubclass(self.model_cls, AbstractSimpleConsensus)

    def test_sampling_funcs(self):
        # print("In test_sampling_funcs for the model {}".format(self.model_cls.name))
        assert self.sampling_funcs  # Not empty
        for sf in self.sampling_funcs:
            assert isinstance(sf, Callable)

    def test_sampling(self, samples):
        # print("In test_sampling for the model {}".format(self.model_cls.name))
        assert isinstance(samples, list)
        assert samples  # Not empty
        for sample in samples:
            assert isinstance(sample, SampleForTest)

    def test_fit_and_compute_consensus(self, samples):
        # print("In test_fit_and_compute_consensus for the model {}".format(self.model_cls.name))
        for sample in samples:
            self._test_fit_and_compute_consensus(sample)


class BaseTestGenerativeConsensusModel(BaseTestSimpleConsensusModel):
    """Abstract test class for generative consensus models.

    Conducts following tests:
    1. `model_cls` class attribute is valid;
    2. `sampling_funcs` class attribute is valid;
    3. Asserts each sample provided by sampling functions is valid;
    4. Runs tests for the following methods of the child consensus model:
        * `fit_and_compute_consensus`,
        * `fit`,
        * `compute_consensus`.

    Notes:
        A child test class only has to override `model_cls` and `sampling_funcs` class attributes.
    """

    # Class of the model to be tested; e.g. StanMultinomialOptimizeConsensus
    model_cls: Type[GenerativeAbstractConsensus] = None

    @classmethod
    def _test_fit(cls, consensus_ref: np.ndarray, parameters_ref: GenerativeAbstractConsensus.Parameters,
                  problem: ConsensusProblem):
        model = cls.model_cls()
        parameters_learned = model.fit(problem, consensus_ref)
        dict_parameters_learned = dataclasses.asdict(parameters_learned)
        dict_parameters_ref = dataclasses.asdict(parameters_ref)
        for p in dict_parameters_ref.keys():
            # print(p, dict_parameters_learned[p], dict_parameters_ref[p])
            log.debug("Distance between learned {p} and reference {p}: {d:f}".format(
                p=p, d=distance(dict_parameters_learned[p], dict_parameters_ref[p])))
            assert close(dict_parameters_learned[p], dict_parameters_ref[p])

    @classmethod
    def _test_compute_consensus(cls, consensus_ref: np.ndarray, parameters_ref: GenerativeAbstractConsensus.Parameters,
                                problem: ConsensusProblem):
        model = cls.model_cls()
        # print("In _test_compute_consensus:", parameters_ref)
        consensus_calculated = model.compute_consensus(problem, parameters_ref)
        if isinstance(consensus_calculated, Tuple):
            consensus_calculated = consensus_calculated[0]
            # TODO (OM, 20210511): Check if the `kwargs_opt` return value of
            #  `AbstractStanOptimizeConsensus.compute_consensus()`is necessary
        assert AllClose.evaluate(consensus_calculated, consensus_ref, atol=cls.ABSOLUTE_TOLERANCE)
        assert 1. - Accuracy.evaluate(problem.f_T, consensus_calculated) <= cls.ABSOLUTE_TOLERANCE

    # ==============================  Actual tests  ==============================
    def test_model_cls(self):  # Override
        # print("In test_model_cls for the model {}".format(self.model_cls.name))
        assert self.model_cls
        assert issubclass(self.model_cls, GenerativeAbstractConsensus)

    def test_fit(self, ref_consensus_params_problem):
        # print("In test_fit for the model {}".format(self.model_cls.name))
        for consensus_ref, parameters_ref, problem, _ in ref_consensus_params_problem:
            self._test_fit(consensus_ref, parameters_ref, problem)

    def test_compute_consensus(self, ref_consensus_params_problem):
        # print("In test_compute_consensus for the model {}".format(self.model_cls.name))
        for consensus_ref, parameters_ref, problem, _ in ref_consensus_params_problem:
            self._test_compute_consensus(consensus_ref, parameters_ref, problem)



from . import log
from .data import Data
from .problems import DiscreteConsensusProblem, JSONDataClass

import dataclasses
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np

DiscreteConsensus = np.ndarray


class AbstractSimpleConsensus:
    """ Base class for very simple consensus algorithms."""
    name = None

    @dataclass
    class Parameters(JSONDataClass):
        """
        Notes:
            This dataclass should be subclassed by each of the different models used to compute consensus
        """
        pass

    def fit_and_compute_consensus(self, dcp: DiscreteConsensusProblem, **kwargs) \
            -> Tuple[DiscreteConsensus, Parameters]:
        raise NotImplementedError

    def fit_and_compute_consensuses_from_data(self, d: Data, questions, **kwargs):
        consensuses = {}
        parameters = {}
        for q in questions:
            consensuses[q], parameters[q] = self.fit_and_compute_consensus_from_data(d, q, **kwargs)
        return consensuses, parameters

    def fit_and_compute_consensus_from_data(self, d: Data, question, **kwargs):
        """Computes consensus and fits model for question question from Data d.

        returns consensus, model parameters"""
        dcp = d.get_dcp(question)
        return self.fit_and_compute_consensus(dcp, **kwargs)


class AbstractConsensus(AbstractSimpleConsensus):
    """ Base class for a consensus algorithm."""
    name = None

    def fit_many_from_data(self, d: Data, reference_consensuses):
        parameters = {}
        for q, consensus in reference_consensuses.items():
            parameters[q] = self.fit_from_data(d, q, consensus)
        return parameters

    def fit_from_data(self, d: Data, question, reference_consensus):
        dcp = d.get_dcp(question)
        return self.fit(dcp, reference_consensus)

    def compute_consensus_from_data(self, d: Data, question, parameters):
        dcp = d.get_dcp(question)
        return self.compute_consensus(dcp, parameters)

    def fit(self, dcp: DiscreteConsensusProblem, reference_consensus: DiscreteConsensus, **kwargs) \
            -> AbstractSimpleConsensus.Parameters:
        """ Fits the model parameters provided that the consensus is already known.
        This is useful to determine the errors of a different set of annotators than the
        ones that were used to determine the consensus.

        returns parameters """
        raise NotImplementedError
    def compute_consensus(self, dcp: DiscreteConsensusProblem, parameters: AbstractSimpleConsensus.Parameters) \
            -> DiscreteConsensus:
        """ Computes the consensus with a fixed pre-determined set of parameters.

        returns consensus """
        raise NotImplementedError


class GenerativeAbstractConsensus(AbstractConsensus):
    """Base class for a consensus algorithm that also samples tasks, workers and annotations."""

    @dataclass
    class DataGenerationParameters(JSONDataClass):
        """
        Notes:
            This dataclass should be subclassed by each of the different generative consensus models
        """
        pass

    def get_dimensions(self, parameters: AbstractSimpleConsensus.Parameters):
        """ Returns the number of labels and number of annotators and number of classes of the model encoded in the
        parameters"""
        raise NotImplementedError

    def sample_tasks(self, dgp: DataGenerationParameters, parameters: Optional[AbstractConsensus.Parameters] = None) \
            -> Tuple[int, np.ndarray]:
        raise NotImplementedError

    def sample_workers(self, dgp: DataGenerationParameters, parameters: Optional[AbstractConsensus.Parameters] = None)\
            -> Tuple[int, Optional[np.ndarray]]:
        raise NotImplementedError

    def sample_annotations(self, tasks, workers, dgp: DataGenerationParameters,
                           parameters: Optional[AbstractConsensus.Parameters] = None)\
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[int]]:
        raise NotImplementedError

    def sample(self, dgp: DataGenerationParameters, parameters: Optional[AbstractConsensus.Parameters] = None):
        n_tasks, tasks = self.sample_tasks(dgp, parameters)
        return self._sample_others(n_tasks, tasks, dgp, parameters)

    def _sample_others(self, n_tasks, tasks, dgp: DataGenerationParameters,
                       parameters: Optional[AbstractConsensus.Parameters] = None):
        n_workers, workers = self.sample_workers(dgp, parameters)
        w_A, t_A, f_A, classes = self.sample_annotations(tasks, workers, dgp, parameters)
        # log.debug(type(w_A.dtype))
        return DiscreteConsensusProblem(n_tasks=n_tasks,
                                        f_T=tasks,
                                        n_workers=n_workers,
                                        w_A=w_A,
                                        t_A=t_A,
                                        f_A=f_A,
                                        classes=classes)

    # TODO: Everything down this comment has to be worked on after freezing the main interfaces.
    # Creates a set of linked discrete consensus problems, with linked meaning that they share the very
    # same set of tasks.
    # Each problem represents how it will be labeled by a different community
    def linked_samples(self, real_parameters, crowds_parameters, dgp: DataGenerationParameters):
        n_tasks, tasks = self.sample_tasks(dgp, parameters=real_parameters)
        crowds_dcps = {}
        for crowd_name, parameters in crowds_parameters.items():
            crowds_dcps[crowd_name] = self._sample_others(n_tasks, tasks, dgp, parameters)

        return tasks, crowds_dcps

    def compute_consensuses(self, crowds_dcps, model, crowd_parameters=None, **kwargs):
        crowds_consensus = {}
        for crowd_name, dcp in crowds_dcps.items():
            # print(dcp)
            # print(model)
            if crowd_parameters is None:
                crowds_consensus[crowd_name], _ = model.fit_and_compute_consensus(dcp, **kwargs)
            else:
                crowds_consensus[crowd_name], _ = \
                    model.fit_and_compute_consensus(dcp, init_params=crowd_parameters[crowd_name], **kwargs)

        return crowds_consensus

    def evaluate_consensuses_on_linked_samples(self, real_parameters, crowds_parameters, models, measures,
                                               dgps: List[DataGenerationParameters],
                                               repeats, init_params=False):
        for dgp in dgps:
            for repetition in range(repeats):
                log.info("..Repetition:", repetition)
                real_labels, crowds_dcps = self.linked_samples(real_parameters, crowds_parameters, dgp)
                for model_name, model in models.items():
                    log.info("...Model:", model_name)
                    crowds_consensus = self.compute_consensuses(crowds_dcps, model,
                                                                crowds_parameters if init_params else None)
                    for measure_name, measure in measures.items():
                        log.info("....Measure:", measure_name)
                        crowds_evals = measure.evaluate_crowds(real_labels, crowds_consensus)
                        for crowd_name, eval_value in crowds_evals.items():
                            d = dataclasses.asdict(dgp)
                            d.update({
                                "consensus_algorithm": model_name,
                                "repetition": repetition,
                                "crowd_name": crowd_name,
                                "measure": measure_name,
                                "value": eval_value})
                            log.info(d)
                            # print(d)
                            yield d



import logging
from logging import NullHandler

log = logging.getLogger(__name__)
log.addHandler(NullHandler())

import crowdnalysis.simple
import crowdnalysis.dawid_skene
import crowdnalysis.cmdstan.multinomial
import crowdnalysis.cmdstan.dawid_skene



from typing import Any, Dict

import numpy as np
import pandas as pd
import pytest
from pandas.api.types import CategoricalDtype

from .conftest import TEST
from ..data import Data
from ..problems import DiscreteConsensusProblem


@pytest.fixture(scope="module")
def dcp_kwargs() -> Dict[str, Dict[str, Any]]:
    kwargs_common = {
        "n_tasks": len(TEST.TASK_IDS),
        "n_workers": np.unique(TEST.USER_IDS).size,
        "t_A": [TEST.TASK_IDS.index(id_) for id_ in TEST.TASK_RUN_IDS],
        "w_A": [np.where(np.unique(TEST.USER_IDS) == id_)[0][0] for id_ in TEST.USER_IDS]}
    kwargs_return = {}
    for q, answers in [(TEST.QUESTIONS[0], TEST.ANSWER_0), (TEST.QUESTIONS[1], TEST.ANSWER_1)]:
        kwargs_q = kwargs_common.copy()
        kwargs_q.update({"n_labels": TEST.CATEGORIES[q].categories.size,
                         "f_A": [TEST.CATEGORIES[q].categories.get_loc(a) for a in answers]})
        kwargs_return[q] = kwargs_q
    # print("\n", kwargs_return)
    return kwargs_return


@pytest.fixture
def mock_pybossa_csv(monkeypatch, fixt_single_file_records):
    """Mocks `pandas.read_csv()` for three different Pybossa files"""
    def mock_read(*args, **kwargs):
        file_name = args[0]
        if file_name == "task.csv":
            records = np.array(TEST.TASK_KEYS, dtype=[("task_key", "U15")])
        elif file_name == "task_info.csv":
            records = np.array(TEST.TASK_IDS, dtype=[(Data.COL_TASK_ID, "i4")])
        else:  # "task_run.csv"
            records = fixt_single_file_records
        df = pd.DataFrame.from_records(records)
        # print("df:\n", df)
        return df
    monkeypatch.setattr(pd, "read_csv", mock_read, raising=True)


@pytest.fixture
def mock_single_file_csv(monkeypatch, fixt_df):
    def mock_read(*args, **kwargs):
        return fixt_df
    monkeypatch.setattr(pd, "read_csv", mock_read, raising=True)


def assert_data_object(d: Data, other_columns=[TEST.EXTRA_COL]):
    """Helper for assertion of a created `Data` object"""
    # First assert the shape
    # asked_questions = set(d.df.columns).intersection(TEST.QUESTIONS)
    asked_questions = set(d.df.columns).intersection(d.questions)
    if other_columns is None:
        other_columns = []
    assert d.df.shape == (len(TEST.TASK_RUN_IDS),
                          len([Data.COL_TASK_ID, Data.COL_USER_ID, *asked_questions, *other_columns,
                               Data.COL_TASK_INDEX, Data.COL_WORKER_INDEX, *[Data.COL_QUESTION_INDEX(q)
                                                                             for q in asked_questions]]))
    # Assert task indices follow the order of the given TEST.TASK_IDS
    assert np.array_equal(d.df[Data.COL_TASK_ID], [d.task_ids[ix] for ix in d.df[Data.COL_TASK_INDEX]])
    # Assert annotator indices follow the order of the sorted unique TEST.USER_IDS
    assert np.array_equal(d.df[Data.COL_USER_ID], [np.unique(TEST.USER_IDS)[ix] for ix in d.df[Data.COL_WORKER_INDEX]])
    # Assert answer indices for a question follow the order of its given TEST.CATEGORIES
    for q in asked_questions:
        assert isinstance(d.df[q].dtype, CategoricalDtype)
        if d.data_src == "test":  # != "test_aidr", because in `Data.from_aidr()` we do not send categories
            assert np.array_equal(d.df[q], [TEST.CATEGORIES[q].categories[ix]
                                            for ix in d.df[Data.COL_QUESTION_INDEX(q)]])


def test_from_df(fixt_data):
    # print("\n", data.df)
    assert_data_object(fixt_data)


def test_from_pybossa(mock_pybossa_csv):
    d = Data.from_pybossa(
            "task_run.csv",
            questions=TEST.QUESTIONS,
            data_src="test",
            preprocess=lambda x: x,
            task_ids=TEST.TASK_IDS,
            categories=TEST.CATEGORIES,
            task_info_file="task_info.csv",
            task_file="task.csv",
            field_task_key="task_key",
            other_columns=[TEST.EXTRA_COL])
    # print("\n", d.df)
    assert_data_object(d)

    # Assert __init__ with None values
    other_columns = None
    d = Data.from_pybossa(
        "task_run.csv",
        questions=None,  # <--
        data_src="test",
        preprocess=lambda x: x,
        task_ids=None,  # <--
        categories=TEST.CATEGORIES,
        task_info_file="task_info.csv",
        task_file="task.csv",
        field_task_key="task_key",
        other_columns=None)  # <--
    # print("\n", d.df)
    assert_data_object(d, other_columns=None)


def test_from_mturk(mock_single_file_csv):
    d = Data.from_mturk(
        "mturk_task.csv",  # dummy value
        questions=TEST.QUESTIONS,
        data_src="test",
        preprocess=lambda x: x,
        task_ids=TEST.TASK_IDS,
        categories=TEST.CATEGORIES,
        other_columns=[TEST.EXTRA_COL])
    # print("\n", d.df)
    assert_data_object(d)

    # Assert __init__ with None values
    d = Data.from_mturk(
        "mturk_task.csv",  # dummy value
        questions=None,  # <--
        data_src="test",
        preprocess=lambda x: x,
        task_ids=None,  # <--
        categories=TEST.CATEGORIES,
        other_columns=None)  # <--
    assert_data_object(d, other_columns=None)


def test_from_aidr(mock_single_file_csv):
    d = Data.from_aidr(
        "aidr.csv",  # dummy value
        questions=[TEST.QUESTIONS[1]],
        data_src="test_aidr",
        preprocess=lambda x: x,
        task_ids=TEST.TASK_IDS,
        other_columns=[TEST.EXTRA_COL])
    # print("\n", d.df)
    assert_data_object(d)

    # Assert __init__ with None values
    d = Data.from_aidr(
        "aidr.csv",  # dummy value
        questions=None,  # <--
        data_src="test_aidr",
        preprocess=lambda x: x,
        task_ids=None,  # <--
        other_columns=None)  # <--
    assert_data_object(d, other_columns=None)


def test_get_categories(fixt_data):
    assert fixt_data.get_categories() == TEST.CATEGORIES


def test_get_dcp(dcp_kwargs, fixt_data):
    for q in TEST.QUESTIONS:
        dcp1 = fixt_data.get_dcp(q)
        # print("\n", dcp_kwargs[q])
        dcp2 = DiscreteConsensusProblem(**dcp_kwargs[q])
        assert dcp1 == dcp2


def test_get_field(fixt_data):
    task_indices = [0, 1]  # -> TEST.TASK_IDS: 12, 10
    for field, answers in [(TEST.QUESTIONS[0], TEST.ANSWER_0), (TEST.QUESTIONS[1], TEST.ANSWER_1)]:
        indices = []
        for ix in task_indices:
            indices += [i for i, x in enumerate(TEST.TASK_RUN_IDS) if x == TEST.TASK_IDS[ix]]
        # print("indices:", indices)
        output = np.array(answers)[indices]
        # print("output:", output)
        # print("return:", data.get_field(task_indices=task_indices, field=field, unique=False))
        assert np.array_equal(output, fixt_data.get_field(task_indices=task_indices, field=field, unique=False))
        assert np.array_equal(np.unique(output),
                              np.unique(fixt_data.get_field(task_indices=task_indices, field=field, unique=True)))


def test_set_classes(fixt_data):
    for q in TEST.QUESTIONS:
        classes_ = TEST.CATEGORIES[q].categories.tolist()
        # Assert a ValueError is raised if the classes is not a sublist of the categories starting from index 0
        with pytest.raises(ValueError):
            fixt_data.set_classes(q, [classes_[0], classes_[-1]])
        assert fixt_data.get_classes(q) == list(range(len(classes_)))
        classes_.pop(-1)
        fixt_data.set_classes(q, classes_)
        assert fixt_data.get_classes(q) == list(range(len(classes_)))
        # Reset the classes
        fixt_data.set_classes(q, None)
        assert fixt_data.get_classes(q) == list(range(len(TEST.CATEGORIES[q].categories.tolist())))


def test_valid_rows(fixt_data):
    # Assert a ValueError is raised if the question is not valid
    with pytest.raises(ValueError):
        fixt_data.valid_rows("-".join(TEST.QUESTIONS))
    # Assert valid rows for non-empty condition
    fixt_data.set_condition(TEST.QUESTIONS[1], "{}=='Yes'".format(TEST.QUESTIONS[0]))
    yes_indices_0 = np.where(np.array(TEST.ANSWER_0) == "Yes")[0]
    assert np.array_equal(yes_indices_0, fixt_data.valid_rows(TEST.QUESTIONS[1]))
    # Assert valid rows for empty condition
    for q in TEST.QUESTIONS:
        fixt_data.set_condition(q, None)
        assert np.array_equal(fixt_data.valid_rows(q), fixt_data.df.index)


def test_set_condition(fixt_data):
    q = TEST.QUESTIONS[1]
    fixt_data.set_condition(q, f"{TEST.QUESTIONS[0]} in ['Yes', 'Not sure']")
    assert np.array_equal(fixt_data.valid_rows(q),
                          [ix for ix, val in enumerate(TEST.ANSWER_0) if val in ["Yes", "Not sure"]])
    # Clear the condition
    fixt_data.set_condition(q, None)
    assert np.array_equal(fixt_data.valid_rows(q), fixt_data.df.index)


def test_make_and_condition():
    assert Data.make_and_condition([]) == ""
    assert Data.make_and_condition([('A', 5), ('B', 'Yes'), ('C C', ['Yes', True])]) \
           == "`A`==5 & `B`=='Yes' & `C C` in ['Yes', True]"
    with pytest.raises(ValueError):
        Data.make_and_condition([('A', 5), ('B', 'Yes'), ('C C', )])


def test_get_others(fixt_data):
    """`Data.[get_tasks, get_workers, get_annotations]`"""
    q = TEST.QUESTIONS[1]
    for method_, col_name in [(fixt_data.get_tasks, Data.COL_TASK_INDEX),
                              (fixt_data.get_workers, Data.COL_WORKER_INDEX),
                              (fixt_data.get_annotations, Data.COL_QUESTION_INDEX(q))]:
        assert np.array_equal(method_(q), fixt_data.df[col_name])
        fixt_data.set_condition(q, f"{TEST.QUESTIONS[0]} in ['Yes', 'Not sure']")
        assert np.array_equal(method_(q),
                          [fixt_data.df[col_name][ix]
                           for ix, val in enumerate(TEST.ANSWER_0) if val in ["Yes", "Not sure"]])
        # Clear the condition
        fixt_data.set_condition(q, None)



import pytest

from ..consensus import AbstractSimpleConsensus
from ..factory import Factory


class DummyConsensus(AbstractSimpleConsensus):
    name = "Dummy"


class NotAnAbstractSimpleConsensus:
    name = "Dummier"


def test_all_factory_methods():
    # Register
    Factory.register_consensus_algorithm(DummyConsensus)
    with pytest.raises(ValueError):
        Factory.register_consensus_algorithm(NotAnAbstractSimpleConsensus)
    # Get
    assert Factory.get_consensus_algorithm(DummyConsensus.name) == DummyConsensus
    with pytest.raises(KeyError):
        Factory.get_consensus_algorithm(NotAnAbstractSimpleConsensus.name)
    # List
    assert DummyConsensus.name in Factory.list_registered_algorithms()
    # Make
    assert isinstance(Factory.make(DummyConsensus.name), DummyConsensus)
    with pytest.raises(KeyError):
        Factory.make(NotAnAbstractSimpleConsensus.name)
    # Unregister
    Factory.unregister_consensus_algorithm(DummyConsensus.name)
    with pytest.raises(KeyError):
        assert Factory.unregister_consensus_algorithm(DummyConsensus.name)



import dataclasses
from .. import log
from typing import List, Tuple, Optional

import numpy as np
from cmdstanpy import CmdStanModel

from ..consensus import GenerativeAbstractConsensus
from dataclasses import dataclass
from ..factory import Factory

from . import AbstractStanOptimizeConsensus
from . import resource_filename

class StanMultinomialOptimizeConsensus(AbstractStanOptimizeConsensus):
    name = "StanMultinomialOptimize"

    @dataclass
    class Parameters(AbstractStanOptimizeConsensus.Parameters):
        tau: np.ndarray = np.array([0.5, 0.5])
        pi: np.ndarray = np.array([[0.9, 0.1], [0.2, 0.8]])

    @dataclass
    class DataGenerationParameters(GenerativeAbstractConsensus.DataGenerationParameters):
        n_tasks: int = 10
        n_annotations_per_task: int = 2

        def __post_init__(self):
            self.n_annotations = self.n_tasks * self.n_annotations_per_task

    def __init__(self):
        AbstractStanOptimizeConsensus.__init__(self, "Multinomial")

    def tau_estimate(self, dcp, ann, k, classes):
        alg = Factory.make("MajorityVoting")
        t_C, _ = alg.fit_and_compute_consensus(dcp)
        single_most_voted = np.argmax(t_C, axis=1)
        unique, counts = np.unique(single_most_voted, return_counts=True)
        # print("u, c:", unique, counts)
        return counts / np.sum(counts)

    def tau_prior(self, dcp, ann, k, classes, alpha=5., beta=5.):
        return alpha * self.tau_estimate(dcp, ann, k, classes) + beta

    def tau_init(self, dcp, ann, k, classes, alpha=0.99):
        beta = (1 - alpha) / k
        return alpha * self.tau_estimate(dcp, ann, k, classes) + beta

    def pi_prior(self, k, l, classes, alpha=1., beta=10):
        labels_which_are_classes = np.zeros((k, l))
        for k_, l_ in enumerate(classes):
            labels_which_are_classes[k_, l_-1] = 1.
        return np.ones((k, l)) * alpha + labels_which_are_classes * beta

    def map_data_to_prior(self, dcp, k, l, classes, ann, **kwargs):
        tau_prior_ = self.tau_prior(dcp, ann, k, classes, alpha=5.)+5.
        #print("tp:", tau_prior_)
        pi_prior_ = self.pi_prior(k, l, classes, alpha=5, beta=5)

        return {"tau_prior": tau_prior_,
                "pi_prior": pi_prior_}

    def map_data_to_inits(self, dcp, ann, k, l, classes, **kwargs):
        pi_prior_ = self.pi_prior(k, l, classes)
        return {'tau': self.tau_init(dcp, ann, k, classes),
                'pi': (pi_prior_ / np.sum(pi_prior_[0]))}

    def map_data_to_args(self, dcp, **kwargs):
        # args = {"iter": 2000}
        return {'algorithm': 'LBFGS',
                'sig_figs': 10,
                #'init_alpha': 0.001
                 'output_dir': "."
                }

    def sample_tasks(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None) \
            -> Tuple[int, Optional[np.ndarray]]:
        model = self.sample_tasks_model()
        sample = model.sample(data={'t': dgp.n_tasks, 'k': parameters.tau.shape[0], 'l': parameters.pi.shape[1]},
                              inits=dataclasses.asdict(parameters),
                              fixed_param=True,
                              iter_sampling=1)
        t_C = sample.stan_variable('t_C').astype(int)[0] - 1
        log.debug("Tasks type: %s", type(t_C.dtype))
        return dgp.n_tasks, t_C

    def sample_workers(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None)\
            -> Tuple[int, Optional[np.ndarray]]:
        return 1, None

    def sample_annotations(self, tasks, workers, dgp: DataGenerationParameters,
                           parameters: Optional[Parameters] = None)\
            -> Tuple[np.ndarray, np.ndarray, np.ndarray, List[int]]:
        model = self.sample_annotations_model()
        sample = model.sample(data={'w': 1,
                                    't': dgp.n_tasks,
                                    'n_annotations_per_task': dgp.n_annotations_per_task,
                                    'k': parameters.tau.shape[0],
                                    'l': parameters.pi.shape[1],
                                    't_C': tasks + 1},
                              inits=dataclasses.asdict(parameters),
                              fixed_param=True,
                              iter_sampling=1)
        t_A = sample.stan_variable('t_A').astype(int)[0]
        t_A -= 1
        w_A = sample.stan_variable('w_A').astype(int)[0]
        log.debug(type(w_A.dtype))
        w_A -= 1
        log.debug(type(w_A.dtype))
        f_A = sample.stan_variable('ann').astype(int)[0]
        f_A -= 1

        return w_A, t_A, f_A, list(range(parameters.tau.shape[0]))


class StanMultinomialEtaOptimizeConsensus(StanMultinomialOptimizeConsensus):
    name = "StanMultinomialEtaOptimize"

    @dataclass
    class Parameters(AbstractStanOptimizeConsensus.Parameters):
        tau: np.ndarray = np.array([0.5, 0.5])
        eta: np.ndarray = np.array([[0.9], [0.8]])
        pi: np.ndarray = np.array([[0.9, 0.1], [0.2, 0.8]])

    def __init__(self):
        AbstractStanOptimizeConsensus.__init__(self, "MultinomialEta")

    def map_data_to_prior(self, dcp, k, l, ann, classes, **kwargs):
        tau_prior_ = self.tau_prior(dcp, ann, k, classes, alpha=5.)
        eta_alpha_prior_ = np.ones((k, l - 1))
        eta_beta_prior_ = np.ones((k, l - 1)) * 0.01
        return {'tau_prior': tau_prior_,
                'eta_alpha_prior': eta_alpha_prior_,
                'eta_beta_prior': eta_beta_prior_}

    def map_data_to_inits(self, dcp, ann, k, l, classes, **kwargs):

        return {'tau': self.tau_init(dcp, ann, k, classes),
                'eta': np.ones((k, l - 1))}

    def compute_consensus_model(self):
        return CmdStanModel(stan_file=resource_filename(StanMultinomialOptimizeConsensus().model_name
                                                        + ".consensus.stan"))
    # def map_data_to_args(self):
    #    #args = {"iter": 2000}
    #    return {'algorithm': 'LBFGS',
    #            'output_dir': "."}

Factory.register_consensus_algorithm(StanMultinomialOptimizeConsensus)
Factory.register_consensus_algorithm(StanMultinomialEtaOptimizeConsensus)


import numpy as np

TOLERANCE = 0.05


def distance(a: np.ndarray, b: np.ndarray):
    return np.linalg.norm(a-b)


def close(a: np.ndarray, b: np.ndarray, eps=TOLERANCE):
    return distance(a, b) < eps



from typing import Tuple

import numpy as np
from statsmodels.stats import inter_rater

from . import log
from .data import Data


def get_n_and_ranges(d: Data, question: str, ignore_le1_annots: bool = True) -> Tuple[np.ndarray, int]:
    """Return 2D (n_tasks, n_labels) annotation count matrix `n`, and # of unique tasks in data `d`.

    Args:
        ignore_le1_annots: Ignores tasks with # of annotations ≤ 1. This makes sense since to speak of an agreement,
            there have to be > 1 annotations for a task.
    """
    dcp = d.get_dcp(question)
    n, _ = dcp.compute_n(ignore_zero_annots=False)
    n = n.sum(axis=0)
    if ignore_le1_annots:
        le1_annots = n.sum(axis=1) <= 1
        if np.any(le1_annots):
            log.warning("{} tasks with ≤ 1 annotations out of {} tasks are ignored in agreement calculation.".format(
                np.sum(le1_annots), n.shape[0]))
            # print("n[le1_annots, :]:\n", n[le1_annots, :])
            n = n[~le1_annots, :]
    return n, dcp.n_tasks


def fleiss_kappa(d: Data, question: str) -> float:
    """Return Fleiss' kappa value for the annotation data

    Raises:
        ValueError: If all tasks don't have the same number of annotations
    """
    n, *_ = get_n_and_ranges(d, question)
    n_sum = np.sum(n, axis=1)
    if np.all(n_sum == np.amax(n_sum)):
        kappa = inter_rater.fleiss_kappa(table=n, method='fleiss')
    else:
        raise ValueError("All tasks must have equal number of annotations")
    return kappa


def gen_fleiss_kappa(d: Data, question: str, w: np.ndarray = None) -> float:
    """Return generalized Fleiss' kappa value for the annotation data

    ref: Gwet KL. (2014) Handbook of Inter-Rater Reliability.
    """
    n, *_ = get_n_and_ranges(d, question)
    # print("\nn:\n", n)
    kappa = _gen_fleiss_kappa(n, w)
    return kappa


def _gen_fleiss_kappa(r, w=None):
    # r[i][l] -> number of raters that assigned item i to category l
    n, q = r.shape
    if w is None:
        w = np.identity(q)

    r_star = np.dot(r, np.transpose(w))
    r_i = np.sum(r, axis=1)
    # print(r_i)
    assert(np.all(r_i > 1))
    den_i = r_i * (r_i-1)
    # print(den_i)
    num_i_k = r * (r_star - 1)
    num_sum_i = np.sum(num_i_k, axis=1)
    # print(num_sum_i)
    p_0 = np.sum(num_sum_i/den_i)/n
    # print("p_0:",p_0)
    pi_k = np.sum(r/r_i[:, np.newaxis], axis=0)/n
    # print(pi_k)
    p_c = np.dot(np.dot(pi_k, w), pi_k)
    return (p_0-p_c)/(1-p_c)


def full_agreement_percentage(d: Data, question: str) -> float:
    """Return the percentage of annotations for the `question` where all annotators agreed on the same answer"""
    n, n_tasks = get_n_and_ranges(d, question)
    best_count = np.amax(n, axis=1)
    pct = np.sum(np.sum(n, axis=1) == best_count) / n_tasks
    return pct



from dataclasses import dataclass

import numpy as np

from crowdnalysis.consensus import AbstractSimpleConsensus
from crowdnalysis.problems import DiscreteConsensusProblem
from .factory import Factory

class Probabilistic(AbstractSimpleConsensus):
    name = "Probabilistic"

    def __init__(self):
        pass

    @classmethod
    def fit_and_compute_consensus(cls, dcp: DiscreteConsensusProblem, softening=0.1):
        n, _ = dcp.compute_n()
        n = n.sum(axis=0) + softening
        n = n[:, dcp.classes]
        consensus = n / np.sum(n, axis=1)[:, np.newaxis]
        # print("Probabilistic._probabilistic_consensus ({}) -> \n".format(consensus.shape), consensus)
        return consensus, AbstractSimpleConsensus.Parameters()


class MajorityVoting(AbstractSimpleConsensus):

    name = "MajorityVoting"

    @dataclass
    class Parameters(AbstractSimpleConsensus.Parameters):
        pass

    def __init__(self):
        pass

    @classmethod
    def consensus_and_single_most_voted(cls, dcp: DiscreteConsensusProblem):
        consensus, _ = Probabilistic.fit_and_compute_consensus(dcp)
        best_index = np.amax(consensus, axis=1)
        bool_best_candidates = (consensus == best_index[:, np.newaxis])
        # A cell is True if the label is a/the best candidate
        num_best_candidates = np.sum(bool_best_candidates, axis=1)
        single_most_voted = np.argmax(consensus, axis=1)
        # TODO: Consider using masked arrays next
        single_most_voted[num_best_candidates != 1] = -1e10  # Mark as -1e20 when there is a tie in the number of votes
        consensus = bool_best_candidates / (num_best_candidates[:, np.newaxis])
        # give probability distribution when num_best_candidates > 1
        # print("MajorityVoting.majority_voting consensus ({}) -> \n{}\n best ({}) -> \n{}".format(consensus.shape,
        # consensus, best.shape, best))
        return consensus, single_most_voted

    def fit_and_compute_consensus(self, dcp: DiscreteConsensusProblem, **kwargs):
        consensus, _ = self.consensus_and_single_most_voted(dcp)
        return consensus, MajorityVoting.Parameters()

Factory.register_consensus_algorithm(MajorityVoting)
Factory.register_consensus_algorithm(Probabilistic)


from typing import Any, List, Optional, Union, Tuple, Callable

import numpy as np
import pandas as pd

from .problems import DiscreteConsensusProblem


class Data:
    """
    This is the main class for storing answers
    - by a set of annotators,
    - to a set of tasks,
    - each task with the very same questions.
    """

    # Column names in Data.df
    COL_USER_ID = "user_id"
    COL_TASK_ID = "task_id"
    COL_TASK_INDEX = "task_index"  #
    COL_WORKER_INDEX = "annotator_index"

    @staticmethod
    def COL_QUESTION_INDEX(question: str) -> str:  # Beware! A constant imposter!
        """Column name for the `question` in `Data.df`"""
        return f"{question}_index"

    def __init__(self):
        """
        The Data object must not be initialized by constructor.
        Instead, use the from methods
        """
        # Make instance attributes' definitions explicit
        self.df = None
        self.data_src = None
        self.questions = None
        self.task_ids = None
        self.task_index_from_id = None
        self.n_tasks = None
        self.annotator_ids = None
        self.annotator_index_from_id = None
        self.n_annotators = None
        self.question_valid_rows = {}
        self._question_classes = {}

    @classmethod
    def from_df(cls, df, data_src=None, task_id_col_name=COL_TASK_ID, annotator_id_col_name=COL_USER_ID,
                questions=None, task_ids=None, categories=None):
        """
        Create a Data object from dataframe df.
        The dataframe contains one row for each annotation,
        each row has a task_id, an annotator id and the answers to a set of
        questions, each answer in one column
        """
        d = Data()

        d.data_src = data_src

        if questions is None:
            questions = list(df.columns)
            questions.remove(task_id_col_name)
            questions.remove(annotator_id_col_name)

        d.set_questions(questions)

        df = df.copy()

        np_task_ids = df[task_id_col_name].to_numpy()
        if task_ids is None:
            task_ids = np.unique(np_task_ids)
        # print(np.unique(np_task_ids)==task_ids)
        d.set_task_ids(task_ids)
        inverse = np.vectorize(d.task_index_from_id.get)(np_task_ids)
        df[cls.COL_TASK_INDEX] = inverse

        # print(df[annotator_id_col_name].to_numpy())
        annotator_ids, inverse = np.unique(df[annotator_id_col_name].to_numpy(), return_inverse=True)
        d.set_annotator_ids(annotator_ids)
        df[cls.COL_WORKER_INDEX] = inverse

        d.set_df(df, categories)

        return d

    @staticmethod
    def make_and_condition(conditions: List[Tuple[str, Union[Any, List[Any]]]]) -> str:
        """Utility function that creates a conjunctive clause from `conditions` to be used in `Data.set_condition()`.

        Args:
            conditions: (column name, value(s)). The value can be a single literal or a list of literals.

        Examples:

            >>> Data.make_and_condition([('info_0', 5), ('info_1', 'Yes'), ('info_2', ['Yes', True])])
            "`info_0`==5 & `info_1`=='Yes' & `info_2` in ['Yes', True]"
        """

        def esc_str(v):
            return v if not isinstance(v, str) else "'{}'".format(v)

        return " & ".join("`{cn}`{op}{v}".format(cn=cn, op="==" if not isinstance(v, list) else " in ",
                                                 v=str(esc_str(v)))
                          for cn, v in conditions)

    def set_condition(self, question: str, conditions: str):
        """Identifies valid rows of `Data.df` for the `question` according to the `conditions`.

        The `question` is asked to an annotator if all conditions are satisfied.

        Args:
            question: Column name for the asked question in the `Data.df` dataframe
            conditions: A valid string to be used in `pandas.DataFrame.query()` that sets the dependency conditions for
                the `question`. Pass "" or None to remove the conditions for the question.

        Examples:
            set_condition("info_3", "`info_0`==5 & `info_1`=='Yes' & `info_2` in ['Yes', True]")

        """
        if conditions:
            self.question_valid_rows[question] = self.df.query(conditions).index
        elif question in self.question_valid_rows:
            del self.question_valid_rows[question]
        # else:  # silently ignore empty conditions

    def _assert_question(self, question: str) -> bool:
        """Helper function

        Raises:
            ValueError: If the `question` is not valid.
        """
        if question not in self.df.columns:
            raise ValueError("{} is not a valid column in dataframe.".format(question))
        return True

    def valid_rows(self, question: str) -> pd.Index:
        """Return the indices of the valid rows for the `question`."""
        self._assert_question(question)
        if question in self.question_valid_rows:
            return self.question_valid_rows[question]
        else:
            return self.df.index

    def set_classes(self, question: str, classes: Optional[List[str]] = None):
        """Specify the `classes` for a `question` which may be different than the label options.

        Raises:
            ValueError: If the `classes` is not a sublist of the `question`'s categories starting from index 0.
        """
        self._assert_question(question)
        if classes is not None:
            cat = self.df[question].dtype
            if classes != cat.categories.tolist()[:len(classes)]:
                raise ValueError("Classes must be a sublist of the question's categories starting from index 0!")
            self._question_classes[question] = [cat.categories.get_loc(x) for x in classes]
        elif question in self._question_classes:
            del self._question_classes[question]
        # else:  # silently ignore already non-existing classes for `question`

    def get_classes(self, question: str) -> List[int]:
        """Return the `classes` for the `question`.

        Label indices are returned if `classes` are not explicitly set.
        """
        self._assert_question(question)
        if question in self._question_classes:
            return self._question_classes[question]
        else:
            return list(range(len(self.df[question].cat.categories)))

    @classmethod
    def _preprocess(cls, df, questions, preprocess=lambda x: x, other_columns=None):
        if other_columns is None:
            other_columns = []
        if questions is None:
            questions = []
        df = df.copy()
        columns_to_retain = [cls.COL_TASK_ID, cls.COL_USER_ID] + questions + other_columns
        df = preprocess(df)
        df = df[columns_to_retain]
        return df

    @classmethod
    def _from_single_file(cls, file_name, questions, data_src=None, preprocess=lambda x: x, task_ids=None,
                          categories=None, other_columns=None, delimiter=","):
        """
        Create a Data object from a file.
        """
        df = pd.read_csv(file_name, delimiter=delimiter)
        # print("_from_single_file -> df columns before preprocessing: ", df.columns)
        df = cls._preprocess(df, questions, preprocess, other_columns)
        # print("_from_single_file -> df columns AFTER preprocessing: ", df.columns)

        return cls.from_df(df, data_src=data_src, annotator_id_col_name=cls.COL_USER_ID, questions=questions,
                           task_ids=task_ids, categories=categories)

    @classmethod
    def from_pybossa(cls, file_name, questions, data_src=None, preprocess=lambda x: x, task_ids=None, categories=None,
                     task_info_file=None, task_file=None, field_task_key="info_media_0", other_columns=None,
                     delimiter=","):
        """
        Create a Data object from a Pybossa file.
        """

        def get_tasks(t_csv, ti_csv, field_task_id, field_task_key):
            t_df = pd.read_csv(t_csv, delimiter=delimiter)
            t_df = t_df[field_task_key]
            ti_df = pd.read_csv(ti_csv, delimiter=delimiter)
            ti_df = ti_df[field_task_id]
            t_df = pd.merge(t_df, ti_df, left_index=True, right_index=True)
            # print("get_tasks {} {} ({}) -> \n{}".format("PyBossa", task_file, len(t_df.index), t_df))
            return t_df

        df = pd.read_csv(file_name, delimiter=delimiter)
        if task_info_file is not None:
            t_df = get_tasks(task_file, task_info_file, cls.COL_TASK_ID, field_task_key)
            df = pd.merge(df, t_df, how="left", left_on=cls.COL_TASK_ID, right_on=cls.COL_TASK_ID)
        # print("from_pybossa -> df columns before preprocessing: ", df.columns)
        df = cls._preprocess(df, questions, preprocess, other_columns=other_columns)
        #  print("from_pybossa -> df columns AFTER preprocessing: ", df.columns)
        # print(df)
        return cls.from_df(df, data_src=data_src, annotator_id_col_name=cls.COL_USER_ID, questions=questions,
                           task_ids=task_ids, categories=categories)

    @classmethod
    def from_mturk(cls, file_name, questions, data_src=None, preprocess=lambda x: x, task_ids=None, categories=None,
                   other_columns=None, delimiter=","):
        """Create a Data object from an Amazon MTurk file."""
        return cls._from_single_file(file_name, questions, data_src, preprocess, task_ids, categories,
                                     other_columns=other_columns, delimiter=delimiter)

    @classmethod
    def from_aidr(cls, file_name, questions, data_src=None, preprocess=lambda x: x, task_ids=None, other_columns=None,
                  delimiter=","):
        """Create a Data object from an AIDR file."""
        # Note: Does NOT send 'categories' arg
        return cls._from_single_file(file_name, questions, data_src, preprocess, task_ids, other_columns=other_columns,
                                     delimiter=delimiter)

    def set_questions(self, questions):
        self.questions = questions

    def set_task_ids(self, task_ids):
        self.task_ids = task_ids
        self.task_index_from_id = {idx: i for i, idx in enumerate(task_ids)}
        self.n_tasks = len(task_ids)

    def set_annotator_ids(self, annotator_ids):
        self.annotator_ids = annotator_ids
        self.annotator_index_from_id = {idx: i for i, idx in enumerate(annotator_ids)}
        self.n_annotators = len(annotator_ids)

    def set_df(self, df, categories=None):
        self.df = df
        for q in self.questions:
            if categories is None:
                cat_values = self.df[q].unique()
                t = pd.api.types.CategoricalDtype(cat_values)
                self.df[q] = self.df[q].astype(t)
            else:
                self.df[q] = self.df[q].astype(categories[q])
        question_indexes = [self.COL_QUESTION_INDEX(x) for x in self.questions]
        self.df[question_indexes] = self.df[self.questions].apply(lambda x: x.cat.codes)

    def n_labels(self, question):
        return len(self.df[question].cat.categories)

    def get_tasks(self, question):
        ix = self.valid_rows(question)
        return self.df.iloc[ix][self.COL_TASK_INDEX].to_numpy()

    def get_workers(self, question):
        ix = self.valid_rows(question)
        return self.df.iloc[ix][self.COL_WORKER_INDEX].to_numpy()

    def get_annotations(self, question):
        ix = self.valid_rows(question)
        return self.df.iloc[ix][self.COL_QUESTION_INDEX(question)].to_numpy()

    # def get_question_matrix(self, question):
    #    df = self.df[[self.COL_TASK_INDEX, self.COL_WORKER_INDEX, question+"_index"]]
    # print(df[question].cat.codes)
    # df[question] = df[question].cat.codes
    #    return df.to_numpy()

    def get_categories(self):
        categories = {}
        for q in self.questions:
            # print(self.df[q].cat)
            # print(type(self.df[q].cat))
            # print(self.df[q].dtype)
            categories[q] = self.df[q].dtype
        return categories

    def get_field(self, task_indices, field, unique=False):
        """Return the `field values` within the `self.df` DataFrame that corresponds to the given `task_indices`

        Args:
            task_indices (np.ndarray): array of integer values on the `task_index` column.
            field (str): `data.df` DataFrame column name
            unique (bool): If True, only unique values will be returned

        Returns:
            np.ndarray: `field` values corresponding to and in the order of the given `task_indices`.
        """
        df_values = self.df[self.df.task_index.isin(task_indices)][[self.COL_TASK_INDEX, field]]  # .isin() loses the order
        df_ind = pd.DataFrame(task_indices, columns=[self.COL_TASK_INDEX])
        df_values = df_ind.merge(df_values, how="left", left_on=self.COL_TASK_INDEX, right_on=self.COL_TASK_INDEX)  # Preserve the order
        df_values = df_values[field]
        if unique:
            df_values = df_values.drop_duplicates()
        return df_values.to_numpy()

    def get_dcp(self, question):
        return DiscreteConsensusProblem(n_tasks=self.n_tasks,
                                        n_workers=self.n_annotators,
                                        t_A=self.get_tasks(question),
                                        w_A=self.get_workers(question),
                                        f_A=self.get_annotations(question),
                                        n_labels=self.n_labels(question),
                                        classes=self.get_classes(question))



import re

from .. import _version as v

# Semantic Versioning Official Regular Expression: https://regex101.com/r/vkijKf/1/
OFFICIAL_SEMVER_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"


def test_version_no():
    assert "__version__" in dir(v)
    assert re.match(OFFICIAL_SEMVER_REGEX, v.__version__)



import dataclasses
import json
from dataclasses import dataclass
from itertools import product
from typing import Dict, List, Optional, Tuple

import numpy as np
from numpyencoder import NumpyEncoder

from . import log


@dataclass
class JSONDataClass:
    def to_json(self):
        return json.dumps(dataclasses.asdict(self), cls=NumpyEncoder)

    @classmethod
    def from_json(cls, s: str):
        d = json.loads(s)
        return cls(**d)

    @classmethod
    def product_from_dict(cls, options: Dict[str, List]):
        keys, values = zip(*options.items())
        ds = [dict(zip(keys, bundle)) for bundle in product(*values)]
        return [cls(**d) for d in ds]


@dataclass(eq=False)
class ConsensusProblem(JSONDataClass):
    """
    Notes: see https://www.mdpi.com/2227-7390/9/8/875, p.4
        W*      A*        T*
        ^        ^        ^
    f_W |    f_A |    f_T |
        |        |        |
        |   w_A  |  t_A   |
        W <----- A -----> T

    """
    n_tasks: int = 0
    # features of the tasks
    f_T: Optional[np.ndarray] = None
    n_workers: int = 0
    # features of the workers
    f_W: Optional[np.ndarray] = None
    n_annotations: Optional[int] = None
    # task of an annotation
    t_A: np.ndarray = np.array([])
    # worker of an annotation
    w_A: np.ndarray = np.array([])
    # features of each of the annotations
    f_A: np.ndarray = np.array([])

    def __post_init__(self):
        if isinstance(self.f_T, list):
            self.f_T = np.array(self.f_T)
        if isinstance(self.f_W, list):
            self.f_W = np.array(self.f_W)
        if isinstance(self.t_A, list):
            self.t_A = np.array(self.t_A)
        if isinstance(self.w_A, list):
            self.w_A = np.array(self.w_A)
        if isinstance(self.f_A, list):
            self.f_A = np.array(self.f_A)
        if len(self.f_A.shape) == 1:
            self.f_A = self.f_A[:, np.newaxis]
        self.n_annotations = self.f_A.shape[0]

    def __eq__(self, other):
        """Compare two `ConsensusProblem`s"""
        # TODO (OM, 20210512): This might need further elaboration depending on the fields of future subclasses
        if not issubclass(self.__class__, other.__class__):
            return False
        return self.to_json() == other.to_json()
        # for field_, val1 in dataclasses.asdict(self).items():
        #     val2 = getattr(other, field_)
        #     if isinstance(val1, np.ndarray):
        #         eq_ = np.array_equal(val1, val2)
        #     else:
        #         eq_ = val1 == val2
        #     if not eq_:
        #         return False
        # return True


@dataclass(eq=False)
class DiscreteConsensusProblem(ConsensusProblem):
    """
    Notes:
        The DiscreteConsensusProblem enforces that:
        - There is a discrete set of real classes
        - The labels should be integers starting from 0.
        - There is a single attribute of each annotation,
          which is also discrete and which at least contains the real_classes.
          It can eventually contain additional answers such as "I do not know", or "in doubt", or "does not apply"

    Example:
        TODO (OM, 20210416): Add examples to init params
    """
    # number of different labels in the annotation
    n_labels: Optional[int] = None
    # Which labels in an annotation correspond to real hidden classes
    classes: Optional[List[int]] = None

    def __post_init__(self):
        ConsensusProblem.__post_init__(self)
        if (self.n_labels is None) and (self.f_A.shape[0] > 0):
            self.n_labels = int(np.max(self.f_A[:, 0]) + 1)
        if self.classes is None:
            # By default every label is a real class
            self.classes = list(range(self.n_labels))

    def __eq__(self, other):
        return super().__eq__(other)

    def compute_n(self, ignore_zero_annots: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the `n` three-dimensional array in Dawid-Skene (1979).

        Dimensions: (n_workers, n_tasks, n_labels)

        Args:
            ignore_zero_annots: If True, filters out tasks with zero # of annotations.

        Returns:
            Tuple of (`n`, indices of filtered tasks)
        """
        # TODO: This should be optimized
        n = np.zeros((self.n_workers, self.n_tasks, self.n_labels))
        for i in range(self.n_annotations):
            n[self.w_A[i], self.t_A[i], self.f_A[i, 0]] += 1
        zero_annots = np.array([])
        if ignore_zero_annots:
            n_sum_w_A = n.sum(axis=0)
            zero_annots = np.all(n_sum_w_A == 0, axis=1)
            if np.any(zero_annots):
                n = n[:, ~zero_annots, :]
                log.warning("{} tasks with zero annotations out of {} tasks in 'n' are eliminated.".format(
                    np.sum(zero_annots), n_sum_w_A.shape[0]))
        return n, np.where(zero_annots)[0]



import numpy as np

from .common import BaseTestSimpleConsensusModel, SampleForTest
from ..dawid_skene import DawidSkene
from ..simple import Probabilistic, MajorityVoting


def sample() -> SampleForTest:
    dgp = DawidSkene.DataGenerationParameters(n_tasks=1000, n_annotations_per_task=200)
    parameters = DawidSkene.Parameters(tau=np.array([0.3, 0.7]), pi=np.array([[[0.6, 0.3, 0.1], [0.3, 0.6, 0.1]]]))
    ds = DawidSkene()
    problem = ds.sample(dgp, parameters)
    return SampleForTest(problem, None)


class TestProbabilisticConsensus(BaseTestSimpleConsensusModel):
    model_cls = Probabilistic
    sampling_funcs = [sample]


class TestMajorityVotingConsensus(BaseTestSimpleConsensusModel):
    model_cls = MajorityVoting
    sampling_funcs = [sample]


from typing import Dict

import numpy as np


class AbstractMeasure:
    @classmethod
    def evaluate(cls, ref_entity: np.ndarray, crowd_entity: np.ndarray, **kwargs) -> float:
        return NotImplementedError

    @classmethod
    def evaluate_crowds(cls, ref_entity, crowd_entities, **kwargs) -> Dict[str, float]:
        return {crowd_entity_name: cls.evaluate(ref_entity, crowd_entity, **kwargs)
                for crowd_entity_name, crowd_entity in crowd_entities.items()}


class AllClose(AbstractMeasure):
    """Check if the ref and compared arrays are element-wise equal within a tolerance."""
    name = "allclose"

    @classmethod
    def evaluate(cls, ref_entity, crowd_entity, atol=0.05) -> bool:
        assert ref_entity.shape == crowd_entity.shape
        return np.allclose(ref_entity, crowd_entity, atol=atol)


class CloseRatio(AbstractMeasure):
    """For the ratio of the number of items that are equal within a tolerance."""
    name = "isclose"

    @classmethod
    def evaluate(cls, ref_entity, crowd_entity, atol=0.05):
        assert ref_entity.shape == crowd_entity.shape
        ratio = np.sum(np.isclose(ref_entity, crowd_entity, atol=atol)) / ref_entity.size
        return ratio


class Accuracy(AbstractMeasure):
    name = "accuracy"

    @classmethod
    def evaluate(cls, real_labels, consensus, **kwargs):
        n_tasks, _ = consensus.shape
        assert real_labels.size == n_tasks
        # real_indexes = real_labels + np.arange(0, I * J, J)
        strict_consensus = np.argmax(consensus, axis=1)
        acc = np.sum(strict_consensus == real_labels) / n_tasks
        return acc


# Accuracy measure taking into account potential label switching
class LSAccuracy(AbstractMeasure):
    name = "LS-accuracy"

    @classmethod
    def evaluate(cls, real_labels, consensus, **kwargs):
        def gen_exchanges():
            swaps = []
            labels = np.unique(real_labels)
            for x in labels:
                for y in range(x):
                    swaps.append((x, y))
            return swaps

        def test_exchanges():
            for x, y in exchanges:
                copy_sc = np.where(strict_consensus == x, -1, strict_consensus)
                copy_sc = np.where(strict_consensus == y, x, copy_sc)
                copy_sc = np.where(copy_sc == -1, y, copy_sc)
                alt_acc = np.sum(copy_sc == real_labels) / n_tasks
                if alt_acc > acc:
                    # print(x, y, "swap OK!")
                    return True, alt_acc, copy_sc[:]
                # else:
                    # print(x, y, "no swap")
            return False, acc, strict_consensus

        n_tasks, _ = consensus.shape
        assert real_labels.size == n_tasks
        # real_indexes = real_labels + np.arange(0, I * J, J)
        strict_consensus = np.argmax(consensus, axis=1)
        acc = np.sum(strict_consensus == real_labels) / n_tasks

        # Now we analyze potential class exchanges one by one

        exchanges = gen_exchanges()
        found, acc, strict_consensus = test_exchanges()
        while found:
            found, acc, strict_consensus = test_exchanges()

        # print("acc=", acc)
        return acc



import numpy as np

from . import log

from .consensus import GenerativeAbstractConsensus, DiscreteConsensusProblem
from .simple import Probabilistic
from .factory import Factory
from dataclasses import dataclass
from typing import Optional, Tuple


class DawidSkene(GenerativeAbstractConsensus):

    name = "DawidSkene"

    @dataclass
    class Parameters(GenerativeAbstractConsensus.Parameters):
        tau: np.ndarray = np.array([0.5, 0.5])
        pi: np.ndarray = np.array([[[0.9, 0.1], [0.2, 0.8]]])

    @classmethod
    def fit_and_compute_consensus(cls, dcp: DiscreteConsensusProblem, max_iterations=10000,
                                  tolerance=1e-3, prior=1.0, verbose=False, init_params=None):
        n, _ = dcp.compute_n()
        # First estimate of T_{i,j} is done by probabilistic consensus
        if init_params is not None:
            tau = init_params.tau
            log_pi = np.log(init_params.pi)
            T, _ = Probabilistic.fit_and_compute_consensus(dcp, softening=prior)
        else:
            T, _ = Probabilistic.fit_and_compute_consensus(dcp, softening=prior)
            # Initialize the percentages of each label
            tau = cls._m_step_p(T, prior)

            # Initialize the errors
            log_pi = cls._m_step_log_pi(T, n, prior)

        has_converged = False
        num_iterations = 0
        any_nan = (np.isnan(T).any() or np.isnan(tau).any() or np.isnan(log_pi).any())
        while (num_iterations < max_iterations and not has_converged) and not any_nan:
            # Expectation step
            old_T = T
            T = cls._e_step(n, log_pi, tau)
            # Maximization
            tau, log_pi = cls._m_step(T, n, prior)
            has_converged = np.allclose(old_T, T, atol=tolerance)
            num_iterations += 1
            any_nan = (np.isnan(T).any() or np.isnan(tau).any() or np.isnan(log_pi).any())
        if any_nan:
            log.info("NaN values detected")
        elif has_converged:
            log.info("DS has converged in %i iterations", num_iterations)
        else:
            log.info("The maximum of %i iterations has been reached", max_iterations)

        return T, cls.Parameters(tau=tau, pi=np.exp(log_pi))

    @classmethod
    def fit(cls, dcp: DiscreteConsensusProblem, T, prior=1.0):
        n, _ = dcp.compute_n()
        tau, log_pi = cls._m_step(T, n, prior)
        return cls.Parameters(tau=tau, pi=np.exp(log_pi))

    @classmethod
    def compute_consensus(cls, dcp: DiscreteConsensusProblem, parameters: Parameters):
        n, _ = dcp.compute_n()
        return cls._e_step(n, np.log(parameters.pi), parameters.tau)

    # Methods from GenerativeAbstractConsensus
    @dataclass
    class DataGenerationParameters(GenerativeAbstractConsensus.DataGenerationParameters):
        n_tasks: int = 10
        n_annotations_per_task: int = 2

        def __post_init__(self):
            self.n_annotations = self.n_tasks * self.n_annotations_per_task

    def get_dimensions(self, parameters: Parameters):
        return parameters.tau.shape[0], parameters.pi.shape[0], parameters.pi.shape[2]

    def sample_tasks(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None):
        if parameters is None:
            parameters = self.Parameters()
        n_classes = len(parameters.tau)  # number of real labels
        # Sample the real labels
        return dgp.n_tasks, np.random.choice(n_classes, size=dgp.n_tasks, p=parameters.tau)

    def sample_workers(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None) \
            -> Tuple[int, Optional[np.ndarray]]:
        return parameters.pi.shape[0], None

    def sample_annotations(self, tasks, workers, dgp: DataGenerationParameters,
                           parameters: Optional[Parameters] = None):
        if parameters is None:
            parameters = self.Parameters()
        n_workers, n_classes, n_labels = parameters.pi.shape
        # Sample the annotators
        annotators = np.random.choice(n_workers, size=(dgp.n_tasks, dgp.n_annotations_per_task))
        labels_and_annotators = annotators + tasks[:, np.newaxis] * n_workers
        labels_and_annotators = labels_and_annotators.flatten()
        unique_la, inverse_la, counts_la = np.unique(labels_and_annotators, return_inverse=True, return_counts=True)
        w_A = np.zeros(dgp.n_annotations, dtype=np.int32)
        f_A = np.zeros(dgp.n_annotations, dtype=np.int32)
        t_A = np.arange(dgp.n_annotations, dtype=np.int32) // dgp.n_annotations_per_task
        for i_la, label_and_annotator in enumerate(unique_la):
            real_label = label_and_annotator // n_workers
            annotator_index = label_and_annotator % n_workers
            emission_p = parameters.pi[annotator_index, real_label]
            emitted_labels = np.random.choice(n_labels, size=counts_la[i_la], p=emission_p)
            ca_indexes = np.equal(inverse_la, i_la)
            w_A[ca_indexes] = annotator_index
            f_A[ca_indexes] = emitted_labels
        classes = list(range(n_classes))
        return w_A, t_A, f_A, classes

    # Private methods

    @classmethod
    def _m_step(cls, T, n, prior):
        return cls._m_step_p(T, prior), cls._m_step_log_pi(T, n, prior)

    @classmethod
    def _m_step_p(cls, T, prior):
        p = np.sum(T, axis=0)
        p += prior
        p /= np.sum(p)
        return p

    @classmethod
    def _m_step_log_pi(cls, T, n, prior):
        log.debug("_m_step_log_pi -> T.shape: {}".format(str(T.shape)))
        log.debug("_m_step_log_pi -> n.shape: {}".format(str(n.shape)))
        _pi = np.swapaxes(np.dot(T.transpose(), n), 0, 1)
        _pi += prior
        sums = np.sum(_pi, axis=2)
        _pi /= sums[:, :, np.newaxis]
        return np.log(_pi)

    @classmethod
    def _e_step(cls, n, log_pi, tau):
        log_T = np.tensordot(n, log_pi, axes=([0, 2], [0, 2]))
        log_T += np.log(tau[np.newaxis, :])
        maxLogT = np.max(log_T)
        logSumT = np.log(np.sum(np.exp(log_T-maxLogT), axis=1)) + maxLogT
        log_T -= logSumT[:, np.newaxis]
        T = np.exp(log_T)
        return T

Factory.register_consensus_algorithm(DawidSkene)


import numpy as np

from crowdnalysis.cmdstan import AbstractStanOptimizeConsensus, resource_filename
from crowdnalysis.cmdstan.multinomial import StanMultinomialOptimizeConsensus, StanMultinomialEtaOptimizeConsensus
from scipy.special import softmax
from cmdstanpy import CmdStanModel
from dataclasses import dataclass
from ..factory import Factory

class StanDSOptimizeConsensus(StanMultinomialOptimizeConsensus):
    name = "StanDSOptimize"

    @dataclass
    class Parameters(AbstractStanOptimizeConsensus.Parameters):
        tau: np.ndarray = np.array([0.5, 0.5])
        pi: np.ndarray = np.array([[[0.9, 0.1], [0.2, 0.8]]])

    def __init__(self):
        AbstractStanOptimizeConsensus.__init__(self, "DS")

    def map_data_to_prior(self, dcp, k, l, classes, ann, **kwargs):
        tau_prior_ = self.tau_prior(dcp, ann, k, classes, 5.)
        pi_prior_ = self.pi_prior(k, l, classes)
        return {"tau_prior": tau_prior_,
                "pi_prior": pi_prior_}

    def map_data_to_inits(self, dcp, ann, k, w, l, classes, **kwargs):
        alg = Factory.make("StanMultinomialEtaOptimize")
        t_C, params = alg.fit_and_compute_consensus(dcp)
        tau_ = params.tau
        multinomial_pi_ = params.pi
        pi_param_ = np.broadcast_to(multinomial_pi_, (w, k, l))
        return {'tau': tau_,
                'pi': pi_param_}

    # TODO: Implement sampling from this model
    sample_tasks = 0
    '''
    @dataclass
    class DataGenerationParameters(StanMultinomialOptimizeConsensus.DataGenerationParameters):
        pass

    def sample_tasks(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None) \
            -> Tuple[int, Optional[np.ndarray]]:
        model = self.sample_tasks_model()
        sample = model.sample(data={'t': dgp.n_tasks, 'k': parameters.tau.shape[0]},
                              inits=dataclasses.asdict(parameters),
                              fixed_param=True,
                              iter_sampling=1)
        t_C = sample.stan_variable('t_C').to_numpy(dtype=int)[0] - 1
        return dgp.n_tasks, t_C

    def sample_workers(self, dgp: DataGenerationParameters, parameters: Optional[Parameters] = None)\
            -> Tuple[int, Optional[np.ndarray]]:
        return 1, None

    def sample_annotations(self, tasks, workers, dgp: DataGenerationParameters, parameters: Optional[Parameters]=None)\
            -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        model = self.sample_annotations_model()
        sample = model.sample(data={'w': 1,
                                    't': dgp.n_tasks,
                                    'n_annotations_per_task': dgp.n_annotations_per_task,
                                    'k': parameters.tau.shape[0],
                                    't_C': tasks + 1},
                              inits=dataclasses.asdict(parameters),
                              fixed_param=True,
                              iter_sampling=1)
        t_A = sample.stan_variable('t_A').to_numpy(dtype=int)[0] - 1
        w_A = sample.stan_variable('w_A').to_numpy(dtype=int)[0] - 1
        f_A = sample.stan_variable('ann').to_numpy(dtype=int)[0] - 1
        return w_A, t_A, f_A
    '''

# class StanDSEtaOptimizeConsensus(StanDSOptimizeConsensus):
#     name = "StanDSEtaOptimize"
#
#     def __init__(self):
#         AbstractStanOptimizeConsensus.__init__(self, "DSEta")
#
#     def map_data_to_prior(self, k, ann, **kwargs):
#         tau_prior_ = self.tau_prior(ann, k, 5.)
#         min_pi_prior_ = np.zeros((k, k - 1))
#         max_pi_prior_ = np.ones((k, k - 1)) * 10
#
#         return {'tau_prior': tau_prior_,
#                 'min_pi_prior': min_pi_prior_,
#                 'max_pi_prior': max_pi_prior_}
#
#     def map_data_to_inits(self, ann, k, w, **kwargs):
#         return {'tau': self.tau_prior(ann, k, alpha=1.),
#                 'eta': np.ones((w, k, k - 1))}
#
#     def map_data_to_args(self, **kwargs):
#         # args = {"iter": 2000}
#         return {'algorithm': 'LBFGS',
#                 'init_alpha': 0.01,
#                 'output_dir': "."}
#

class StanDSEtaHOptimizeConsensus(StanDSOptimizeConsensus):
    name = "StanDSEtaHOptimize"

    def __init__(self):
        AbstractStanOptimizeConsensus.__init__(self, "DSEtaH")

    def map_data_to_prior(self, dcp, k, l, ann, classes, **kwargs):
        tau_prior_ = self.tau_prior(dcp, ann, k, classes, alpha=5.)
        eta_alpha_prior_ = np.ones((k, l - 1))
        eta_beta_prior_ = np.ones((k, l - 1)) * 0.01
        return {'tau_prior': tau_prior_,
                'eta_alpha_prior': eta_alpha_prior_,
                'eta_beta_prior': eta_beta_prior_}

    def compute_movements(self, eta, classes):
        k, l = eta.shape
        l += 1
        dst = np.zeros((k,l-1), dtype=np.int16)
        for _k in range(k):
            for _l in range(l-1):
                dst[_k][_l] = _l + (_l >= (classes[_k]-1));
        return dst

    def softmax_diag(self, eta, classes):
        dst = self.compute_movements(eta, classes)
        k, l = eta.shape
        l += 1
        pi = np.zeros((k, l))
        for _k in range(k):
            pi[_k][dst[_k]] = -eta[_k]
            pi[_k] = softmax(pi[_k])
        return pi

    def map_data_to_inits(self, dcp, ann, k, w, l, classes, **kwargs):
        eta = np.ones((k, l - 1))*5.
        print("eta:", eta)
        pi_prior_ = self.pi_prior(k, l, classes)
        print("old_pi_prior:", pi_prior_)
        pi_prior_ = self.softmax_diag(eta, classes)
        print("pi_prior:", pi_prior_)
        pi_param_ = np.broadcast_to(pi_prior_ / np.sum(pi_prior_[0]), (w, k, l))
        print("param:", pi_param_)
        ti = self.tau_init(dcp, ann, k, classes)
        print("tau_init =", ti)
        return {'tau': ti,
                'eta': eta,
                'pi': pi_param_}

    def map_data_to_args(self, dcp, **kwargs):
        # args = {"iter": 2000}
        return {'algorithm': 'LBFGS',
                #'init_alpha': 0.01,
                'sig_figs': 10,
                'output_dir': "."}

    def compute_consensus_model(self):
        return CmdStanModel(stan_file=resource_filename(StanDSOptimizeConsensus().model_name
                                                        + ".consensus.stan"))

Factory.register_consensus_algorithm(StanDSOptimizeConsensus)
# Factory.register_consensus_algorithm(StanDSEtaOptimizeConsensus)
Factory.register_consensus_algorithm(StanDSEtaHOptimizeConsensus)


import numpy as np
import pandas as pd
import pytest
from pandas.api.types import CategoricalDtype

from ..data import Data


class TEST:
    TASK_IDS = [12, 10, 11, 13]  # Task 13 has no annotations
    TASK_KEYS = [f"t{i}" for i in TASK_IDS]
    TASK_RUN_IDS = [10, 11, 12, 11, 12, 11, 10, 10, 12, 10, 11, 12]
    USER_IDS = ["u2", "u0", "u1", "u1", "u2", "u2", "u0", "u1", "u0", "u3", "u3", "u3"]
    QUESTIONS = ["question_0", "question_1"]
    CATEGORIES = {QUESTIONS[0]: CategoricalDtype(categories=["Yes", "No", "Not sure"], ordered=False),
                  QUESTIONS[1]: CategoricalDtype(categories=["A", "B", "C", "Not answered"], ordered=False)}
    ANSWER_0 = ["No", "No", "Yes", "Yes", "Not sure", "Yes", "Not sure", "Yes", "No", "No", "Yes", "No"]
    ANSWER_1 = ["C", "Not answered", "B", "B", "C", "A", "B", "A", "B", "Not answered", "A", "A"]
    EXTRA_COL = "extra_col"
    EXTRA_COL_VAL = [f"img_{id_}" for id_ in TASK_RUN_IDS]


@pytest.fixture(scope="session")
def fixt_single_file_records() -> np.ndarray:
    return np.array([(TEST.USER_IDS[i], TEST.TASK_RUN_IDS[i], TEST.ANSWER_0[i], TEST.ANSWER_1[i], TEST.EXTRA_COL_VAL[i])
                     for i in range(len(TEST.TASK_RUN_IDS))],
                    dtype=[(Data.COL_USER_ID, "U15"), (Data.COL_TASK_ID, "i4"),
                           (TEST.QUESTIONS[0], "U15"), (TEST.QUESTIONS[1], "U15"), (TEST.EXTRA_COL, "U15")])


@pytest.fixture(scope="session")
def fixt_df(fixt_single_file_records):
    df = pd.DataFrame.from_records(fixt_single_file_records)
    return df


@pytest.fixture(scope="session")
def fixt_data_factory(fixt_df):
    class DataFactory:
        @classmethod
        def make(cls, data_src="test"):
            d = Data.from_df(fixt_df, data_src=data_src, questions=TEST.QUESTIONS, task_ids=TEST.TASK_IDS,
                             categories=TEST.CATEGORIES, annotator_id_col_name=Data.COL_USER_ID)
            # print("\n", d.df)
            return d
    return DataFactory


@pytest.fixture(scope="session")
def fixt_data(fixt_data_factory) -> Data:
    d = fixt_data_factory.make()
    # print("\n", d.df)
    return d




from html.parser import HTMLParser
from unittest.mock import mock_open, patch
# TODO (OM, 20210702): Replace unittest with pytest, if a solid way of mocking `builtins.open` is found for the latter.

import pandas as pd
import pytest

from .conftest import TEST
from ..factory import Factory
from ..visualization import csv_description, html_description, html_description_crossed


question = TEST.QUESTIONS[0]
picture_field = TEST.EXTRA_COL


class HTMLParserForTest(HTMLParser):

    def __init__(self):
        self.start_tags = []
        self.end_tags = []
        super().__init__()

    def handle_starttag(self, tag, attrs):
        self.start_tags.append(tag)

    def handle_endtag(self, tag):
        self.end_tags.append(tag)


def assert_valid_html(html: str) -> HTMLParserForTest:
    """Asserts if a given string is valid html"""
    parser = HTMLParserForTest()
    parser.feed(html)

    assert (parser.start_tags[0].lower() == "html")
    assert (parser.end_tags[-1].lower() == "html")
    assert (sorted(parser.start_tags) == sorted(parser.end_tags))  # all tags are closed, valid XHTML

    return parser


@pytest.fixture(scope="module")
def fixt_consensus(fixt_data):
    model = Factory.make("MajorityVoting")
    consensus, _ = model.fit_and_compute_consensus(fixt_data.get_dcp(question))
    return consensus


def test_html_description(fixt_consensus, fixt_data):
    mocked_file_path = "mock/file/path/spam.html"
    with patch("builtins.open", mock_open()) as mocked_file:
        html = html_description(consensus=fixt_consensus, data=fixt_data, question=question,
                                picture_field=picture_field, output_file=mocked_file_path)
        # assert if opened file on write mode 'w'
        mocked_file.assert_called_once_with(mocked_file_path, 'w')
        # assert if the html content was written in file
        mocked_file().write.assert_called_once_with(html)
    # assert generated HTML is valid
    assert_valid_html(html)


def test_html_description_crossed(fixt_consensus, fixt_data):
    compare_task_indices = list(range(len(TEST.TASK_IDS)))[1:]
    mocked_file_path = "mock/file/path/spam_crossed.html"
    with patch("builtins.open", mock_open()) as mocked_file:
        html = html_description_crossed(consensus=fixt_consensus, data=fixt_data, question=question,
                                        picture_field=picture_field, compare_task_indices=compare_task_indices,
                                        lbl_outline="spam", lbl_actor="eggs", output_file=mocked_file_path)
        # assert if opened file on write mode 'w'
        mocked_file.assert_called_once_with(mocked_file_path, 'w')
        # assert if the html content was written in file
        mocked_file().write.assert_called_once_with(html)
    # assert generated HTML is valid
    parser = assert_valid_html(html)

    html2 = html_description(consensus=fixt_consensus, data=fixt_data, question=question, picture_field=picture_field)
    parser2 = HTMLParserForTest()
    parser2.feed(html2)
    # assert `html` has more tags than `html2` due to `compare_task_indices`
    assert(len(parser.start_tags) > len(parser2.start_tags))


def test_csv_description(fixt_consensus, fixt_data):
    mocked_file_path = "mock/file/path/spam.csv"
    with patch.object(pd.DataFrame, "to_csv") as mocked_file:
        csv = csv_description(consensus=fixt_consensus, data=fixt_data, question=question,
                              picture_field=picture_field, output_file=mocked_file_path)
        # assert if the html content was written in file
        mocked_file.assert_called_with(mocked_file_path)



import dataclasses

import numpy as np
import pandas as pd
import pytest

from . import close
from .conftest import TEST
from .. import analysis, factory
from ..measures import Accuracy


@pytest.fixture(scope="module")
def d_expert(fixt_data_factory):  # see conftest.py
    d = fixt_data_factory.make(data_src="test_expert")
    return d


@pytest.fixture(scope="module")
def d_others(fixt_data_factory):
    d_others_ = {}
    for i in range(2):
        data_src = f"test_crowd_{i}"
        d_others_[data_src] = fixt_data_factory.make(data_src=data_src)
    # print([d.data_src for d in d_others_.values()])
    return d_others_


@pytest.fixture(scope="module")
def expert_consensuses_n_parameters(d_expert):
    model_ = factory.Factory.make("DawidSkene")
    expert_consensuses, expert_parameters = model_.fit_and_compute_consensuses_from_data(d_expert, TEST.QUESTIONS)
    return expert_consensuses, expert_parameters


@pytest.fixture(scope="module")
def parameters_others(expert_consensuses_n_parameters, d_others):
    model_ = factory.Factory.make("DawidSkene")
    expert_consensuses, _ = expert_consensuses_n_parameters
    parameters_others_ = analysis.compute_crossed(model_, d_others, expert_consensuses)
    return parameters_others_


def test_compute_crossed(expert_consensuses_n_parameters, parameters_others):
    _, expert_parameters = expert_consensuses_n_parameters
    for question, dgp in expert_parameters.items():
        expert_dgp = dataclasses.asdict(dgp)
        for param_name, expert_param_val in expert_dgp.items():
            for other in parameters_others:
                other_dgp = dataclasses.asdict(parameters_others[other][question])
                assert np.array_equal(expert_param_val, other_dgp[param_name])


def test_prospective_analysis(d_expert, expert_consensuses_n_parameters, parameters_others):
    _, expert_parameters = expert_consensuses_n_parameters
    measures = {Accuracy.name: Accuracy}
    repeats = 2
    question = TEST.QUESTIONS[0]
    model_ = factory.Factory.make("DawidSkene")
    ds = factory.Factory.make("DawidSkene")
    mv = factory.Factory.make("MajorityVoting")
    models = {"DS": ds, "MV": mv}
    options = {"n_tasks": [10, 100],
               "n_annotations_per_task": [3, 9]}
    scenarios = ds.DataGenerationParameters.product_from_dict(options)
    df = analysis.prospective_analysis(question, d_expert.data_src, expert_parameters, parameters_others,
                                       generative_model=model_, models=models, measures=measures, dgps=scenarios,
                                       repeats=repeats)
    # print(df)
    assert isinstance(df, pd.DataFrame)
    assert df.shape[0] == repeats * len(models) * len(options["n_tasks"]) * \
           len(options["n_annotations_per_task"]) * (len(parameters_others.keys()) + 1)  # +1 for experts


def test_compare_data_to_consensus(d_expert, d_others, expert_consensuses_n_parameters):
    expert_consensuses, _ = expert_consensuses_n_parameters
    for d_other in d_others.values():
        for question in TEST.QUESTIONS:
            df = analysis.compare_data_to_consensus(d_expert, d_other, expert_consensuses, question,
                                                    add_total_cols=True)
            # assert df is DataFrame
            assert isinstance(df, pd.DataFrame)
            # #Rows == #Labels
            assert df.shape[0] == d_other.n_labels(question)
            # #Columns == #Classes
            assert (df.shape[1] - 1 == len(d_other.get_classes(question)))
            # Total column == Sum(other columns)
            assert np.allclose(df.iloc[:, -1].to_numpy(), df.iloc[:, :-1].sum(axis=1).to_numpy())


def test_gen_confusion_matrix(d_expert, expert_consensuses_n_parameters):
    expert_consensuses, _ = expert_consensuses_n_parameters
    for question in TEST.QUESTIONS:
        df = analysis.gen_confusion_matrix(expert_consensuses[question], expert_consensuses[question],
                                           d_expert, question)
        # assert df is DataFrame
        assert isinstance(df, pd.DataFrame)
        # assert all categories are in the index and in the columns
        categories = d_expert.get_categories()[question].categories.tolist()
        assert df.index.to_list() == categories
        assert df.columns.to_list() == categories
        # assert dataframe total sum = # of tasks
        assert close(np.nansum(df.to_numpy()), len(np.unique(d_expert.get_tasks(question))))
        # TODO (OM, 20210716): assert the values of rows?



import numpy as np

from .. import log
from .common import BaseTestGenerativeConsensusModel, SampleForTest
from ..cmdstan.multinomial import StanMultinomialOptimizeConsensus, StanMultinomialEtaOptimizeConsensus


def easy_sample() -> SampleForTest:
    dgp = StanMultinomialOptimizeConsensus.DataGenerationParameters(n_tasks=1000, n_annotations_per_task=20)
    parameters = StanMultinomialOptimizeConsensus.Parameters(tau=np.array([0.3, 0.7]),
                                                             pi=np.array([[0.9, 0.1], [0.2, 0.8]]))
    smoc = StanMultinomialOptimizeConsensus()
    problem = smoc.sample(dgp, parameters)
    return SampleForTest(problem, parameters)


def sample_non_finite_gradient() -> SampleForTest:
    dgp = StanMultinomialOptimizeConsensus.DataGenerationParameters(n_tasks=100000, n_annotations_per_task=3)
    parameters = StanMultinomialOptimizeConsensus.Parameters(tau=np.array([0.05, 0.95]),
                                                             pi=np.array([[0.999, 0.001], [0.2, 0.8]]))
    smoc = StanMultinomialOptimizeConsensus()
    problem = smoc.sample(dgp, parameters)
    log.info("sample_non_finite_gradient")
    log.info(problem)
    log.info(parameters)
    return SampleForTest(problem, parameters)


def more_labels_than_classes_sample() -> SampleForTest:
    dgp = StanMultinomialOptimizeConsensus.DataGenerationParameters(n_tasks=1000, n_annotations_per_task=20)
    parameters = StanMultinomialOptimizeConsensus.Parameters(tau=np.array([0.3, 0.7]),
                                                                pi=np.array([[0.6, 0.1, 0.3], [0.3, 0.5, 0.2]]))
    smoc = StanMultinomialOptimizeConsensus()
    problem = smoc.sample(dgp, parameters)
    return SampleForTest(problem, parameters)


class TestStanMultinomialOptimizeConsensus(BaseTestGenerativeConsensusModel):
    model_cls = StanMultinomialOptimizeConsensus
    sampling_funcs = [easy_sample, sample_non_finite_gradient, more_labels_than_classes_sample]


class TestStanMultinomialEtaOptimizeConsensus(BaseTestGenerativeConsensusModel):
    model_cls = StanMultinomialEtaOptimizeConsensus
    sampling_funcs = [easy_sample, sample_non_finite_gradient, more_labels_than_classes_sample]




from typing import List, Type


#from . import dawid_skene, simple
from .consensus import AbstractSimpleConsensus
#from .cmdstan import multinomial as sm
#from .cmdstan import dawid_skene as sds
#from crowdnalysis.cmdstan.multinomial import StanMultinomialOptimizeConsensus

class Factory:
    """Factory class for consensus algorithms"""
    _algorithms = {}

    @classmethod
    def _msg_exception(cls, name: str, options: bool = True) -> str:
        """Return the message to be used in `KeyError`s"""
        msg = "{} algorithm is not registered.".format(name)
        if options:
            msg = "{}. Available options are {}.".format(msg, str(cls.list_registered_algorithms()))
        return msg

    @classmethod
    def make(cls, name: str, **kwargs) -> AbstractSimpleConsensus:
        """Return an instance of the algorithm registered with the specified `name` and created by the `kwargs`.

        Raises:
            KeyError: If the algorithm is not registered.

        """
        try:
            return cls._algorithms[name](**kwargs)
        except KeyError:
            raise KeyError(cls._msg_exception(name, options=True))

    @classmethod
    def get_consensus_algorithm(cls, name: str) -> Type[AbstractSimpleConsensus]:
        """Return the corresponding consensus algorithm class (not its instance)

        Raises:
            KeyError: If the algorithm is not registered.

        """
        try:
            return cls._algorithms[name]
        except KeyError:
            raise KeyError(cls._msg_exception(name, options=True))

    @classmethod
    def register_consensus_algorithm(cls, algorithm: AbstractSimpleConsensus) -> None:
        """Register a new consensus algorithm.

        Raises:
            ValueError: If `algorithm` is not a `AbstractConsensus`

        """
        if not issubclass(algorithm, AbstractSimpleConsensus):
            raise ValueError(f"{str(algorithm)} is not a 'AbstractConsensus' subclass.")
        cls._algorithms[algorithm.name] = algorithm
        return None

    @classmethod
    def unregister_consensus_algorithm(cls, name: str) -> None:
        """Unregisters an existing consensus algorithm.

        Raises:
            KeyError: If the algorithm is not registered.

        """
        try:
            del cls._algorithms[name]
        except KeyError:
            raise KeyError(cls._msg_exception(name, options=False))
        return None

    @classmethod
    def list_registered_algorithms(cls) -> List[str]:
        """Return the list of registered algorithm names"""
        return list(cls._algorithms.keys())








from .common import BaseTestGenerativeConsensusModel
from ..cmdstan.dawid_skene import StanDSOptimizeConsensus, StanDSEtaHOptimizeConsensus

from .test_dawid_skene import sample

class TestStanDSOptimizeConsensus(BaseTestGenerativeConsensusModel):
    model_cls = StanDSOptimizeConsensus
    sampling_funcs = [sample]

#class TestStanDSEtaOptimizeConsensus(BaseTestGenerativeConsensusModel):
#    model_cls = StanDSEtaOptimizeConsensus
#    sampling_funcs = [sample]

class TestStanDSEtaHOptimizeConsensus(BaseTestGenerativeConsensusModel):
    model_cls = StanDSEtaHOptimizeConsensus
    sampling_funcs = [sample]




import pytest

import numpy as np

from . import TOLERANCE
from ..measures import Accuracy, AllClose, CloseRatio, LSAccuracy


@pytest.fixture(scope="module")
def ref_consensus():
    consensus = np.array([[0.61, 0.17, 0.18, 0.04],   # 0
                          [0.25, 0.17, 0.40, 0.18],   # 2
                          [0.02, 0.64, 0.16, 0.18],   # 1
                          [0.02, 0.51, 0.18, 0.29],   # 1
                          [0.23, 0.14, 0.28, 0.35]])  # 3
    return consensus


@pytest.fixture(scope="module")
def real_labels():
    return np.array([0, 2, 1, 1, 3])  # see `ref_consensus`


def test_accuracies_binary():
    real = np.array([0, 1, 0, 1, 0])
    consensus_all = np.array([[0.8, 0.2], [0.1, 0.7], [0.6, 0.4], [0.45, 0.55], [0.99, 0.01]])
    # Switch 0<->1
    consensus_none = np.ones(consensus_all.shape) - consensus_all

    assert Accuracy.evaluate(real, consensus_all) == 1.0
    assert Accuracy.evaluate(real, consensus_none) == 0.0
    assert LSAccuracy.evaluate(real, consensus_all) == 1.0
    assert LSAccuracy.evaluate(real, consensus_none) == 1.0


def test_accuracies_multilabel(real_labels, ref_consensus):
    real2 = np.array([0, 2, 1, 2, 3])  # different than the consensus
    # Switch labels 1<->3
    consensus_switched = np.array([[i[0], i[3], i[2], i[1]] for i in ref_consensus.tolist()])
    assert Accuracy.evaluate(real_labels, ref_consensus) == 1.0       # [0, 2, 1, 1, 3] vs [0, 2, 1, 1, 3]
    assert Accuracy.evaluate(real_labels, consensus_switched) == 0.4  # [0, 2, 1, 1, 3] vs [0, 2, 3, 3, 1]
    assert LSAccuracy.evaluate(real_labels, ref_consensus) == 1.0
    assert LSAccuracy.evaluate(real_labels, consensus_switched) == 1.0

    assert Accuracy.evaluate(real2, ref_consensus) == 0.8       # [0, 2, 1, 2, 3] vs [0, 2, 1, 1, 3]
    assert Accuracy.evaluate(real2, consensus_switched) == 0.4  # [0, 2, 1, 2, 3] vs [0, 2, 3, 3, 1]
    assert LSAccuracy.evaluate(real2, ref_consensus) == 0.8
    assert LSAccuracy.evaluate(real2, consensus_switched) == 0.8

    # Switch labels 0<->3, 1<->2
    consensus_switched = np.array([[i[3], i[2], i[1], i[0]] for i in ref_consensus.tolist()])
    assert Accuracy.evaluate(real_labels, ref_consensus) == 1.0       # [0, 2, 1, 1, 3] vs [0, 2, 1, 1, 3]
    assert Accuracy.evaluate(real_labels, consensus_switched) == 0.0  # [0, 2, 1, 1, 3] vs [3, 1, 2, 2, 0]
    assert LSAccuracy.evaluate(real_labels, ref_consensus) == 1.0
    assert LSAccuracy.evaluate(real_labels, consensus_switched) == 1.0

    assert Accuracy.evaluate(real2, ref_consensus) == 0.8       # [0, 2, 1, 2, 3] vs [0, 2, 1, 1, 3]
    assert Accuracy.evaluate(real2, consensus_switched) == 0.2  # [0, 2, 1, 2, 3] vs [3, 1, 2, 2, 0]
    assert LSAccuracy.evaluate(real2, ref_consensus) == 0.8
    assert LSAccuracy.evaluate(real2, consensus_switched) == 0.8


def test_closeratio(ref_consensus):
    assert CloseRatio.evaluate(ref_consensus, ref_consensus, atol=0.) == 1.0
    assert CloseRatio.evaluate(ref_consensus, ref_consensus - TOLERANCE, atol=TOLERANCE) == 1.0
    assert CloseRatio.evaluate(ref_consensus, ref_consensus - 2 * TOLERANCE, atol=TOLERANCE) == 0.0
    n_labels = ref_consensus.shape[1]
    compared = np.copy(ref_consensus)
    compared[:, 0] -= 2 * TOLERANCE
    assert CloseRatio.evaluate(ref_consensus, compared) == (n_labels - 1.) / n_labels


def test_allclose(ref_consensus):
    assert AllClose.evaluate(ref_consensus, ref_consensus, atol=0.)
    assert AllClose.evaluate(ref_consensus, ref_consensus - TOLERANCE, atol=TOLERANCE)
    assert not AllClose.evaluate(ref_consensus, ref_consensus - 2 * TOLERANCE)



import numpy as np

from .common import BaseTestGenerativeConsensusModel, BaseTestSimpleConsensusModel, SampleForTest
from ..dawid_skene import DawidSkene


def sample() -> SampleForTest:
    dgp = DawidSkene.DataGenerationParameters(n_tasks=1000, n_annotations_per_task=20)
    parameters = DawidSkene.Parameters(tau=np.array([0.3, 0.7]), pi=np.array([
        [[0.8, 0.1, 0.1], [0.1, 0.8, 0.1]],
        [[0.5, 0.2, 0.3], [0.2, 0.5, 0.3]]
    ]))
    ds = DawidSkene()
    problem = ds.sample(dgp, parameters)
    return SampleForTest(problem, parameters)


def sample_w_kwargs() -> SampleForTest:
    sample_ = sample()
    return SampleForTest(sample_.problem, None, {"init_params": sample_.parameters})


class TestDawidSkeneConsensus(BaseTestGenerativeConsensusModel):
    model_cls = DawidSkene
    sampling_funcs = [sample]


class TestDawidSkeneConsensusSimple(BaseTestSimpleConsensusModel):
    # Test `fit_and_compute_consensus` only with `init_params` arg
    model_cls = DawidSkene
    sampling_funcs = [sample_w_kwargs]



import os
from typing import List, Union

import numpy as np
import pandas as pd
from dominate import document, tags, util

from .data import Data


IMG_ON_ERROR = "this.src='{}';".format("https://pbs.twimg.com/profile_images/1260593552515641345/j_TGePnL_400x400.jpg")


def html_description(consensus: np.ndarray, data: Data, question: str, picture_field: str, width=120, height=90,
                     dec=3, warn_threshold=.1, output_file: str = None, pretty=True) -> str:
    """Returns an HTML string that displays the images of tasks.

    Optionally, saves the string to an HTML file.

    Args:
        consensus: Consensus probabilities of tasks
        data : Annotation data
        question: Question for annotation
        picture_field: column name in the `Data.df` dataframe
        width: max image width
        height: max image height
        dec: number of decimal digits of probabilities to display
        warn_threshold: threshold difference between best and second best consensus probabilities to be marked
            as warning
        output_file: (optional) full path to the output HTML file
        pretty: True, for a more human readable HTML string; False, for a smaller sized file.

    Returns:
        HTML string

    """

    FRAME_COLOR = "#ef0707"
    STYLE = """
        .labels {{
          overflow-x: scroll;
          overflow-y: hidden;
          white-space: nowrap;
          -webkit-overflow-scrolling: touch;
        }}
        .task-image {{
          display: inline-block;
        }}
        .task-image img {{
          max-width: {w}px;
          max-height: {h}px;
        }}
        .warn{{
          padding:2px;
          border:6px solid {c};
        }}
    """.format(w=width, h=height, c=FRAME_COLOR)
    NOTES = ("- Hover on any image to read its best and second best consensus probabilities.<br/>"
             "- When these two probabilities have a difference &le; {t}, the image is framed with "
             "<span style='color:{c}; font-weight:bold'>borders</span> and marked as warning.<br/>"
             "- Scroll horizontally to view all images.").format(t=warn_threshold, c=FRAME_COLOR)
    TITLE = "{} Consensus :: {}".format(data.data_src, question.title())

    best = np.argmax(consensus, axis=1)
    labels = np.unique(best)
    label_names = list(data.df[question].cat.categories)
    dcp = data.get_dcp(question)
    if dcp.n_tasks > consensus.shape[0]:
        _, filtered_tasks = dcp.compute_n(ignore_zero_annots=True)
    else:
        filtered_tasks = []
    
    with document(title=TITLE) as doc:
        with doc.head:
            tags.style(STYLE)
            tags.base(target="_blank")
        with doc.body as body:
            body.add(tags.h1(TITLE, style="color:Tomato; text-align:center"))
            body.add(tags.p(util.raw(NOTES)))
            for label in labels:
                task_indices = np.where(best == label)[0]
                data_task_indices = task_indices
                for i in filtered_tasks:  # Increment tasks indices accordingly to reflect filtered tasks
                    data_task_indices = np.apply_along_axis(lambda x: np.where(x >= i, x + 1, x), axis=0,
                                                            arr=data_task_indices)
                task_picture_links = data.get_field(data_task_indices, picture_field, unique=True)
                label_id = "label" + str(label)
                body.add(tags.h2("{} ({}):".format(str(label_names[label]).title(), len(task_indices)), id=label_id))
                n_warn = 0
                with body.add(tags.div(cls='labels')):
                    for ix, tpl in enumerate(task_picture_links):
                        ti = tags.div(cls="task-image")
                        task_index = task_indices[ix]
                        probabilities = [(p, li) for li, p in enumerate(consensus[task_index, :])]
                        probabilities.sort(key=lambda x: x[0], reverse=True)
                        probs_str = "\n".join("- {0}: {1:.{2}f}".format(
                            label_names[li], p, dec) for p, li in probabilities[:2])
                        img_title = "Task Index: {}\n{}".format(str(data_task_indices[ix]), probs_str)
                        if probabilities[0][0] - probabilities[1][0] <= warn_threshold:
                            n_warn += 1
                            kwargs = {"cls": "warn"}
                        else:
                            kwargs = {}
                        ti += tags.a(tags.img(src=tpl, title=img_title, onerror=IMG_ON_ERROR, **kwargs), href=tpl)
                    if n_warn > 0:
                        label_header = doc.body.getElementById(label_id)
                        label_header.add_raw_string(" (<span style='color:{c}'>{n}</span> warning{s})".format(
                            c=FRAME_COLOR, n=str(n_warn), s="s" if n_warn > 1 else ""))
            # Filtered tasks
            if len(filtered_tasks):
                label_id = "label_filtered"
                body.add(tags.h2("Tasks without annotations ({}):".format(len(filtered_tasks)), id=label_id))
                filtered_task_picture_links = data.get_field(filtered_tasks, picture_field, unique=True)
                with body.add(tags.div(cls='labels')):
                    for ix, tpl in enumerate(filtered_task_picture_links):
                        ti = tags.div(cls="task-image")
                        img_title = "Task Index: {}".format(str(filtered_tasks[ix]))
                        ti += tags.a(tags.img(src=tpl, title=img_title, onerror=IMG_ON_ERROR), href=tpl)
    html_str = doc.render(pretty=pretty, xhtml=True)
    if output_file:
        with open(output_file, 'w') as f:
            f.write(html_str)
        print("HTML for the {} consensus for the question '{}' is saved into file:\n '{}'".format(
            data.data_src, question, os.path.relpath(output_file)))
    return html_str


def csv_description(consensus: np.ndarray, data: Data, question: str, picture_field: str, dec=3,
                    output_file: str = None) -> Union[None, str]:
    """Saves/returns the CSV-format representation of the consensus on tasks.

    `diff_best_two` column is the difference between the best and second best consensus probabilities.

    Args:
        consensus: Consensus probabilities of tasks
        data : Annotation data
        question: Question for annotation
        picture_field: column name in the `Data.df` dataframe
        dec: number of decimal digits of probabilities to display
        output_file: (optional) full path to the output HTML file

    Returns:
        If output_file is None, returns the resulting csv format as a string. Otherwise, returns None.
    """

    def diff_best_two(x):
        probabilities = list(x)
        probabilities.sort(reverse=True)
        return probabilities[0] - probabilities[1]

    label_names = list(data.df[question].cat.categories)
    df = pd.DataFrame(consensus, columns=label_names)
    df["diff_best_two"] = df.apply(diff_best_two, axis=1)
    df.index.name = "task_index"
    df[picture_field] = data.get_field(list(df.index), picture_field, unique=True)
    df = round(df, dec)
    val = df.to_csv(output_file)
    if output_file:
        print("CSV-format output for the {} consensus for the question '{}' is saved into file:\n '{}'".format(
            data.data_src, question, os.path.relpath(output_file)))
    return val


def html_description_crossed(consensus: np.ndarray, data: Data, question: str, compare_task_indices: List[int],
                             picture_field: str, lbl_outline: str, lbl_actor: str, included: bool = True,
                             width=120, height=90, dec=3, warn_threshold=.1,
                             output_file: str = None, pretty=True) -> str:
    """Return an HTML string that displays the images of tasks. Outline the tasks which are [in/not in] the given list.

    Optionally, saves the string into an HTML file.

    Args:
        consensus: Consensus probabilities of tasks
        data : Annotation data
        question: Question for annotation
        compare_task_indices: List of task indices of to be compared with the tasks in the `consensus`
        picture_field: column name in the `Data.df` dataframe
        lbl_outline: Label for the tasks whose `included` comparison results `True`
        lbl_actor: The component who generated the `compare_task_indices` (e.g. POLIMI's pipeline)
        included: If True, outlines the `consensus` tasks `ìn` the `compare_task_indices`;
            otherwise, outlines `not in` tasks
        width: max image width
        height: max image height
        dec: number of decimal digits of probabilities to display
        warn_threshold: threshold difference between best and second best consensus probabilities to be marked
            as warning
        output_file: (optional) full path to the output HTML file. If omitted, file is not saved.
        pretty: True, for a more human readable HTML string; False, for a smaller sized file.

    Returns:
        HTML string

    """
    # TODO (OM, 20210115): Refactor to remove duplicated code inside 'html_description'?

    FRAME_COLOR = "#ef0707"
    OUTLINE_COLOR = "orange"
    STYLE = """
        .labels {{
          overflow-x: scroll;
          overflow-y: hidden;
          white-space: nowrap;
          -webkit-overflow-scrolling: touch;
        }}
        .task-image {{
          padding: 2px;
          display: inline-block;
        }}
        .task-image img {{
          max-width: {w}px;
          max-height: {h}px;
        }}
        .warn{{
          padding: 2px;
          border: 6px solid {c};
        }}
        .outline{{
          outline: 3px solid {o};
        }}
        .summary{{
          color: white;
          text-align: center;
          outline: 2px solid #626262;
          outline-style: double;
          background-color: #8c8c8c;
        }}
    """.format(w=width, h=height, c=FRAME_COLOR, o=OUTLINE_COLOR)
    NOTES = ("- Hover on any image to read its best and second best consensus probabilities.<br/>"
             "- When these two probabilities have a difference &le; {t}, the image is framed with "
             "<span style='color:{c}; font-weight:bold'>borders</span> and marked as warning.<br/>"
             "- Otherwise, the image is regarded as 'good' (regardless of whether it is relevant or not).<br/>"
             "- Images which were <b>{m}</b> by <em>{a}</em> are "
             "<span style='color:{o}; font-weight:bold'>outlined</span>.<br/>"
             "- Scroll horizontally to view all images.").format(t=warn_threshold, c=FRAME_COLOR, o=OUTLINE_COLOR,
                                                                 a=lbl_actor, m=lbl_outline.lower())
    TITLE = "{} Consensus :: {}".format(data.data_src, question.title())

    def label_info(n_warn_, n_outline_, msg_action):
        info_ = ""
        if n_warn_ > 0:
            info_ = "<span style='color:{c}'>{n}</span> warning{s}".format(
                c=FRAME_COLOR, n=str(n_warn_), s="s" if n_warn_ > 1 else "")
        if n_outline_ > 0:
            if info_:
                info_ += ", "
            info_ += "<span style='color:{o}'>{n}</span> {m}".format(
                o=OUTLINE_COLOR, n=str(n_outline_), m=msg_action.lower())
        info_ = " ({})".format(info_)
        return info_

    def make_summary(consensus_, data_, n_warn_total_, n_outline_good_total_, n_outline_warn_total_, msg_action):
        str_summary_ = ("Consensus Warnings: <span style='color:{c}'>{tw:.1%}</span><br/>"
                        "{m} Warnings: <span style='color:{o}'>{tdw:.1%}</span>, "
                        "{m} Good: <span style='color:{o}'>{tdg:.1%}</span>").format(
            c=FRAME_COLOR, o=OUTLINE_COLOR, tw=n_warn_total_ / consensus_.shape[0],
            tdw=0 if n_warn_total_ == 0 else n_outline_warn_total_ / n_warn_total_,
            tdg=n_outline_good_total_ / (data_.n_tasks - n_warn_total_),
            m=msg_action.capitalize())
        return str_summary_

    best = np.argmax(consensus, axis=1)
    labels = np.unique(best)
    label_names = list(data.df[question].cat.categories)
    dcp = data.get_dcp(question)
    if dcp.n_tasks > consensus.shape[0]:
        _, filtered_tasks = dcp.compute_n(ignore_zero_annots=True)
    else:
        filtered_tasks = []

    with document(title=TITLE) as doc:
        with doc.head:
            tags.style(STYLE)
            tags.base(target="_blank")
        with doc.body as body:
            body.add(tags.h1(TITLE, style="color:Tomato; text-align:center"))
            body.add(tags.p(util.raw(NOTES)))
            summary_id = "summary"
            body.add(tags.h1("", id=summary_id, cls="summary"))
            n_warn_total = 0
            n_outline_warn_total = 0
            n_outline_good_total = 0
            for label in labels:
                task_indices = np.where(best == label)[0]
                data_task_indices = task_indices
                for i in filtered_tasks:  # Increment tasks indices accordingly to reflect filtered tasks
                    data_task_indices = np.apply_along_axis(lambda x: np.where(x >= i, x + 1, x), axis=0,
                                                            arr=data_task_indices)
                task_picture_links = data.get_field(data_task_indices, picture_field, unique=True)
                label_id = "label" + str(label)
                body.add(tags.h2("{} ({}):".format(str(label_names[label]).title(), len(task_indices)), id=label_id))
                n_warn = 0
                n_outline_warn = 0
                n_outline_good = 0
                with body.add(tags.div(cls='labels')):
                    for ix, tpl in enumerate(task_picture_links):
                        ti = tags.div(cls="task-image")
                        task_index = task_indices[ix]
                        for i in filtered_tasks:  # Increment tasks indices accordingly to reflect filtered tasks
                            task_indices = np.apply_along_axis(lambda x: np.where(x >= i, x + 1, x), axis=0,
                                                               arr=task_indices)
                        probabilities = [(p, li) for li, p in enumerate(consensus[task_index, :])]
                        probabilities.sort(key=lambda x: x[0], reverse=True)
                        probs_str = "\n".join("- {0}: {1:.{2}f}".format(
                            label_names[li], p, dec) for p, li in probabilities[:2])
                        img_title = "Task Index: {}\n{}".format(str(data_task_indices[ix]), probs_str)
                        is_warning = probabilities[0][0] - probabilities[1][0] <= warn_threshold
                        is_outlined = (included == (task_index in compare_task_indices))
                        if is_warning:
                            n_warn += 1
                            kwargs = {"cls": "warn"}
                        else:
                            kwargs = {}
                        if is_outlined:
                            if is_warning:
                                n_outline_warn += 1
                            else:
                                n_outline_good += 1
                            kwargs["cls"] = (kwargs.get("cls", "") + " outline").lstrip()
                        ti += tags.a(tags.img(src=tpl, title=img_title, **kwargs), href=tpl)
                    n_outline = n_outline_warn + n_outline_good
                    if n_warn + n_outline > 0:
                        label_header = doc.body.getElementById(label_id)
                        label_header.add_raw_string(label_info(n_warn, n_outline, lbl_outline))
                n_warn_total += n_warn
                n_outline_warn_total += n_outline_warn
                n_outline_good_total += n_outline_good
            summary_elm = doc.body.getElementById(summary_id)
            summary_elm.add_raw_string(make_summary(consensus, data, n_warn_total,
                                                    n_outline_good_total, n_outline_warn_total, lbl_outline))

    html_str = doc.render(pretty=pretty, xhtml=True)
    if output_file:
        with open(output_file, 'w') as f:
            f.write(html_str)
        print("HTML for the {s} '{q}' consensus vs {a}-{m} tasks is saved into:\n'{f}'".format(
            s=data.data_src, q=question, a=lbl_actor, m=lbl_outline, f=os.path.relpath(output_file)))
    return html_str



"""Module for analysing crowd-sourced data"""

from typing import List

import numpy as np
import pandas as pd

from .import data
from .consensus import AbstractSimpleConsensus, GenerativeAbstractConsensus


def compute_crossed(model, d_others, ref_consensuses):
    """Compute parameters when the true labels are known

    Args:
        model (consensus.AbstractConsensus):
        d_others (Dict[str, data.Data]): Dictionary of (data source name, Data) key-value pairs
        ref_consensuses (Dict[str, np.ndarray]): Dictionary of (question, consensus) key-value pairs

    Returns:
        Dict[str, Dict[str, AbstractSimpleConsensus.Parameters]]: {data_source: {question: {_p: [...], _pi: [...]}}}
    """
    parameters_others = {}
    for d_name in d_others:
        parameters_others[d_name] = model.fit_many_from_data(d_others[d_name], ref_consensuses)
    return parameters_others


def compare_data_to_consensus(d_base, d_compare, base_consensuses, question, add_total_cols=False):
    """Compares the cross comparison of a crowd-sourced data with the consensus

    Args:
        d_base (data.Data): `Data` used to generate the `base_consensuses
        d_compare (data.Data): `Data` to be compared
        base_consensuses (Dict[str, np.ndarray]): Dictionary of (question, consensus) key-value pairs
        question (str): The question to be compared. e.g. "severity"
        add_total_cols (bool): If True, adds checksum columns.

    Returns:
        pd.DataFrame
    """
    consensus_col_labels = dict(zip(list(range(base_consensuses[question].shape[1])),
                                    d_base.get_categories()[question].categories.tolist()))
    # print("consensus_col_labels:", consensus_col_labels)
    df_consensus = pd.DataFrame(base_consensuses[question]).rename(columns=consensus_col_labels)
    df_consensus[data.Data.COL_TASK_INDEX] = list(range(df_consensus.shape[0]))
    df_out = df_consensus.merge(d_compare.df[[question, data.Data.COL_TASK_INDEX]], how="right",
                                left_on=data.Data.COL_TASK_INDEX, right_on=data.Data.COL_TASK_INDEX)
    df_out.drop([data.Data.COL_TASK_INDEX], axis=1, inplace=True)
    df_out = df_out.groupby([question]).sum()
    if add_total_cols:
        # df_out["Total Consensus"] = df_out.sum(axis=1)
        df_count_compare = pd.DataFrame(d_compare.df[question], columns=[question])
        df_out["Total"] = df_count_compare.groupby([question]).size()
    return df_out


def prospective_analysis(question, expert_data_src, expert_parameters, parameters_others, generative_model, models,
                         measures, dgps: List[GenerativeAbstractConsensus.DataGenerationParameters], repeats=2):
    """Makes a predictive analysis for each community based on the expert parameters.

    `repeats` times analysis is made for each model, for each no of tasks, for each no of annotations per task.

    Args:
        question (str): The question to be compared. e.g. "severity"
        expert_data_src (str): Data source of the experts' data
        expert_parameters (Dict[str, np.ndarray]): {_p: [...], _pi: [...]}
        parameters_others (Dict[str, Dict[str, np.ndarray]]): {data_source: {_p: [...], _pi: [...]}}:
        generative_model (consensus.GenerativeAbstractConsensus):
        # TODO (OM, 20210304): When GenerativeAbstractConsensus methods are converted to class/static methods
        # this argument can be omitted.
        models (Dict[str, consensus.AbstractConsensus]]: Dictionary of (model_abbreviation, model_instance) pairs
        measures (Dict[str, measures.AbstractMeasure]): Dictionary of (measure name, measure class) pairs
        dgps: List of data generation parameters
        repeats (int): Number of times the analysis should be repeated

    Returns:
        pd.DataFrame.
    """
    crowds_parameters = {name: parameters_others[name][question] for name in parameters_others}
    crowds_parameters[expert_data_src] = expert_parameters[question]
    # print(expert_parameters[question])
    return pd.DataFrame.from_records(
        generative_model.evaluate_consensuses_on_linked_samples(
            expert_parameters[question], crowds_parameters, models, measures, dgps, repeats=repeats))


def gen_confusion_matrix(consensus_ref: np.ndarray, consensus_compare: np.ndarray,
                         d_ref: data.Data, question: str) -> pd.DataFrame:
    """Compute confusion matrix for a consensus based on another ref consensus.

    Cells values are the number of tasks. The values are merely the sum of probabilities for the corresponding
    (class, label) pair.
    """
    best_ref = np.argmax(consensus_ref, axis=1)
    categories = d_ref.get_categories()[question].categories.tolist()
    label_names = dict(zip(range(len(categories)), categories))
    # print("label_names:", label_names)
    df_out = pd.DataFrame(consensus_compare).rename(columns=label_names)
    df_out["Ground Truth"] = best_ref
    df_out = df_out.groupby("Ground Truth").sum()
    df_out = df_out.rename(index=label_names)
    df_out = df_out.reindex(categories)  # In case there are labels which were not in `best_ref`
    return df_out



from typing import Tuple

import numpy as np
import pandas as pd
import pytest

from . import close
from .. import agreement
from ..data import Data


class BaseTestDataSimple:
    TASK_IDS = None
    TASK_RUN_IDS = None
    USER_IDS = None
    ANSWERS = None
    QUESTION = None


class TEST_SIMPLE(BaseTestDataSimple):
    TASK_IDS = [0, 1, 2, 3, 4]
    TASK_RUN_IDS = [0, 0, 0, 1, 1, 1, 2, 2, 3]  # task 2 has two, task 3 has one, task 4 has zero annotations
    USER_IDS = ["u0", "u1", "u2", "u0", "u1", "u2", "u0", "u1", "u0"]
    # USER_IDS = [0, 1, 2, 0, 1, 2, 0, 1, 0]
    ANSWERS = ["No", "No", "No", "No", "No", "Yes", "No", "Yes", "No"]
    QUESTION = "question"


class _TEST_KAPPA(BaseTestDataSimple):
    pass


def _make_data(test_data: BaseTestDataSimple) -> Data:
    records = np.array([(test_data.USER_IDS[i], test_data.TASK_RUN_IDS[i], test_data.ANSWERS[i])
                        for i in range(len(test_data.TASK_RUN_IDS))],
                       dtype=[(Data.COL_USER_ID, "U15"), (Data.COL_TASK_ID, "i4"), (test_data.QUESTION, "U15")])
    df = pd.DataFrame.from_records(records)
    d = Data.from_df(df, task_ids=test_data.TASK_IDS, questions=["question"], annotator_id_col_name=Data.COL_USER_ID)
    # print("\n", d.df)
    return d


@pytest.fixture(scope="module")
def data_():
    d = _make_data(TEST_SIMPLE)
    # print("\nData simple:\n", d.df)
    return d


@pytest.fixture(scope="module")
def data_kappa() -> Tuple[BaseTestDataSimple, str, float]:
    """Table for test data copied from Wikipedia

    see https://en.wikipedia.org/wiki/Fleiss%27_kappa#Worked_example

    Returns:
        Tuple: (annotation data, question, kappa value)
    """
    table = np.asarray("""\
        0   0   0   0   14
        0   2   6   4   2
        0   0   3   5   6
        0   3   9   2   0
        2   2   8   1   1
        7   7   0   0   0
        3   2   6   3   0
        2   5   3   2   2
        6   5   2   1   0
        0   2   2   3   7""".split(), float).reshape(10, 5)
    kappa = 0.210

    num_ann_per_task = np.max(np.sum(table, axis=1))
    _TEST_KAPPA.TASK_IDS = np.arange(table.shape[0])
    _TEST_KAPPA.TASK_RUN_IDS = np.array([np.repeat(i, num_ann_per_task) for i in _TEST_KAPPA.TASK_IDS]).flatten()
    _TEST_KAPPA.USER_IDS = np.array([np.arange(num_ann_per_task) for _ in _TEST_KAPPA.TASK_IDS], dtype="U15").flatten()
    _TEST_KAPPA.ANSWERS = np.array([])
    for i in range(table.shape[0]):
        for j in range(table.shape[1]):
            a = np.repeat(j, table[i, j])
            if a.size > 0:
                _TEST_KAPPA.ANSWERS = np.append(_TEST_KAPPA.ANSWERS, a)
    _TEST_KAPPA.QUESTION = "question"
    d = _make_data(_TEST_KAPPA)
    # print("\nData kappa:\n", d.df)
    return d, _TEST_KAPPA.QUESTION, kappa


def test_get_n_ranges(data_):
    n, n_tasks = agreement.get_n_and_ranges(data_, TEST_SIMPLE.QUESTION, ignore_le1_annots=True)
    assert n_tasks == len(TEST_SIMPLE.TASK_IDS)
    assert np.array_equal(n, np.array([[3, 0], [2, 1], [1, 1]]))
    n, _n_tasks = agreement.get_n_and_ranges(data_, TEST_SIMPLE.QUESTION, ignore_le1_annots=False)
    assert n_tasks == len(TEST_SIMPLE.TASK_IDS)
    assert np.array_equal(n, np.array([[3, 0], [2, 1], [1, 1], [1, 0], [0, 0]]))


def test_full_agreement(data_):
    assert agreement.full_agreement_percentage(data_, TEST_SIMPLE.QUESTION) == .2  # Only for task 0.


def test_fleiss_kappa(data_, data_kappa):
    with pytest.raises(ValueError):  # All tasks must have equal number of annotations for standard Fleiss' kappa
        agreement.fleiss_kappa(data_, TEST_SIMPLE.QUESTION)
    # Assert kappa
    d, q, kappa = data_kappa
    assert close(agreement.fleiss_kappa(d, q), kappa, eps=0.001)


def test_gen_fleiss_kappa(data_, data_kappa):
    # TODO (OM, 20210521): Manually calculate the kappa and ensure that it is ~ -0.385
    assert close(agreement.gen_fleiss_kappa(data_, TEST_SIMPLE.QUESTION), -0.385, eps=0.001)
    d, q, kappa = data_kappa
    assert close(agreement.gen_fleiss_kappa(d, q), kappa, eps=0.001)


#!/usr/bin/env python

import os
import re

from setuptools import setup, find_packages


PACKAGE_NAME = "crowdnalysis"
HERE = os.path.dirname(__file__)
PACKAGE_DIR = "src"

CLASSIFIERS = """
Programming Language :: Python :: 3
License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)
Operating System :: OS Independent
Intended Audience :: Science/Research
Topic :: Scientific/Engineering :: Information Analysis
"""


def readme() -> str:
    with open(os.path.join(HERE, 'README.md'), 'r') as f:
        content = f.read()
    return content


def requirements() -> str:
    with open(os.path.join(HERE, "requirements.txt"), "r") as f:
        content = f.read()
    return content.splitlines()


def get_version() -> str:
    version_file = open(os.path.join(HERE, PACKAGE_DIR, PACKAGE_NAME, "_version.py"))
    version_contents = version_file.read()
    return re.search("__version__ = \"(.*?)\"", version_contents).group(1)


setup(
    name=PACKAGE_NAME,
    version=get_version(),
    author="IIIA-CSIC",
    description="Library to help analyze crowdsourcing results",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/Crowd4SDG/crowdnalysis",
    project_urls={"Bug Tracker": "https://github.com/Crowd4SDG/crowdnalysis/issues"},
    install_requires=requirements(),
    classifiers=CLASSIFIERS.strip().split('\n'),
    package_dir={"": PACKAGE_DIR},
    packages=find_packages(where=PACKAGE_DIR),
    package_data={PACKAGE_NAME: ['cmdstan/cmdstan/*.stan']},
    python_requires=">=3.7"
)
